# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from os.path import isfile
from platform import system

from pytest import mark, raises

from ffpuppet import BrowserTerminatedError, BrowserTimeoutError, FFPuppet

from .puppet_target import PuppetTarget
from .target import Target, TargetLaunchError, TargetLaunchTimeout

def test_puppet_target_01(mocker, tmp_path):
    """test creating a PuppetTarget"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.reason = FFPuppet.RC_CLOSED
    fake_ffp.return_value.log_length.return_value = 562
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), None, 300, 25, 5000, None, 25) as target:
        assert target.closed
        assert target._browser_logs is None
        assert isfile(target.prefs)
        prefs_file = target.prefs
        assert target._tmp_prefs
        assert target.detect_failure([], False) == Target.RESULT_NONE
        assert target.log_size() == 1124
        fake_ffp.return_value.log_length.assert_any_call("stderr")
        fake_ffp.return_value.log_length.assert_any_call("stdout")
        assert target.monitor is not None
        target.add_abort_token("test")
        assert fake_ffp.return_value.add_abort_token.call_count == 1
        target.save_logs("fake_dest")
        assert fake_ffp.return_value.save_logs.call_count == 1
    assert fake_ffp.return_value.clean_up.call_count == 1
    assert not isfile(prefs_file)
    # with extra args
    with PuppetTarget(str(fake_file), None, 1, 1, 1, None, 1, rr=True, fake=1) as target:
        pass

def test_puppet_target_02(mocker, tmp_path):
    """test PuppetTarget.launch()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    # test providing prefs.js
    with PuppetTarget(str(fake_file), None, 300, 25, 5000, str(fake_file), 35) as target:
        assert target.prefs == str(fake_file)
        assert not target._tmp_prefs
        target.launch("launch_target_page")
        assert target._browser_logs is None
        assert fake_ffp.return_value.launch.call_count == 1
        assert fake_ffp.return_value.close.call_count == 0
        fake_ffp.return_value.launch.side_effect = BrowserTimeoutError
        with raises(TargetLaunchTimeout):
            target.launch("launch_target_page")
        assert fake_ffp.return_value.launch.call_count == 2
        assert fake_ffp.return_value.close.call_count == 1
        fake_ffp.return_value.launch.side_effect = BrowserTerminatedError
        with raises(TargetLaunchError):
            target.launch("launch_target_page")

def test_puppet_target_03(mocker, tmp_path):
    """test PuppetTarget.detect_failure()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.RC_ALERT = FFPuppet.RC_ALERT
    fake_ffp.RC_CLOSED = FFPuppet.RC_CLOSED
    fake_ffp.RC_EXITED = FFPuppet.RC_EXITED
    fake_ffp.RC_WORKER = FFPuppet.RC_WORKER
    fake_file = tmp_path / "fake"
    fake_file.touch()
    target = PuppetTarget(str(fake_file), None, 300, 25, 5000, str(fake_file), 25)
    # no failures
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.reason = None
    assert target.detect_failure([], False) == Target.RESULT_NONE
    assert target.detect_failure(["memory"], False) == Target.RESULT_NONE
    assert not target.closed
    fake_ffp.reset_mock()
    # test close
    fake_ffp.return_value.is_healthy.return_value = False
    fake_ffp.return_value.is_running.return_value = False
    fake_ffp.return_value.reason = FFPuppet.RC_CLOSED
    assert target.detect_failure([], False) == Target.RESULT_NONE
    assert fake_ffp.return_value.is_running.call_count == 1
    assert fake_ffp.return_value.is_healthy.call_count == 1
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    # test single process crash
    fake_ffp.return_value.is_healthy.return_value = False
    fake_ffp.return_value.is_running.return_value = False
    fake_ffp.return_value.reason = FFPuppet.RC_ALERT
    assert target.detect_failure([], False) == Target.RESULT_FAILURE
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    # test multiprocess crash
    fake_ffp.return_value.is_healthy.return_value = False
    fake_ffp.return_value.is_running.return_value = True
    fake_ffp.return_value.reason = FFPuppet.RC_ALERT
    assert target.detect_failure([], False) == Target.RESULT_FAILURE
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    # test exit with no crash logs
    fake_ffp.return_value.is_healthy.return_value = False
    fake_ffp.return_value.is_running.return_value = False
    fake_ffp.return_value.reason = FFPuppet.RC_EXITED
    assert target.detect_failure([], False) == Target.RESULT_NONE
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    # test timeout
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.is_running.return_value = True
    fake_ffp.return_value.reason = None
    fake_ffp.return_value.cpu_usage.return_value = ((1234, 10), (1236, 75), (1238, 60))
    fake_kill = mocker.patch("grizzly.target.puppet_target.kill", autospec=True)
    assert target.detect_failure([], True) == Target.RESULT_FAILURE
    if system() == "Linux":
        assert fake_kill.call_count == 1
        assert fake_ffp.return_value.wait.call_count == 1
    else:
        assert fake_kill.call_count == 0
        assert fake_ffp.return_value.wait.call_count == 0
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    # test timeout ignored
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.is_running.return_value = True
    fake_ffp.return_value.reason = None
    assert target.detect_failure(["timeout"], True) == Target.RESULT_IGNORED
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    # test worker
    fake_ffp.return_value.is_healthy.return_value = False
    fake_ffp.return_value.is_running.return_value = False
    fake_ffp.return_value.reason = FFPuppet.RC_WORKER
    assert target.detect_failure([], False) == Target.RESULT_FAILURE
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    # test memory ignored
    fake_ffp.return_value.is_healthy.return_value = False
    fake_ffp.return_value.is_running.return_value = False
    fake_ffp.return_value.reason = FFPuppet.RC_WORKER
    fake_ffp.return_value.available_logs.return_value = " ffp_worker_memory_usage "
    assert target.detect_failure(["memory"], False) == Target.RESULT_IGNORED
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    # test log-limit ignored
    fake_ffp.return_value.is_healthy.return_value = False
    fake_ffp.return_value.is_running.return_value = False
    fake_ffp.return_value.reason = FFPuppet.RC_WORKER
    fake_ffp.return_value.available_logs.return_value = " ffp_worker_log_size "
    assert target.detect_failure(["log-limit"], False) == Target.RESULT_IGNORED
    assert fake_ffp.return_value.close.call_count == 1
    fake_ffp.reset_mock()
    # test browser closing test case
    fake_ffp.return_value.is_healthy.return_value = False
    fake_ffp.return_value.is_running.return_value = False
    fake_ffp.return_value.reason = FFPuppet.RC_EXITED
    target.forced_close = False
    target.rl_countdown = 0
    assert target.detect_failure([], False) == Target.RESULT_NONE
    assert fake_ffp.return_value.wait.call_count == 1
    assert fake_ffp.return_value.close.call_count == 1

@mark.skipif(system() == "Windows", reason="Unsupported on Windows")
def test_puppet_target_04(mocker, tmp_path):
    """test PuppetTarget.dump_coverage()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_proc = mocker.patch("grizzly.target.puppet_target.Process", autospec=True)
    fake_proc.return_value.children.return_value = (mocker.Mock(pid=101),)
    fake_proc_iter = mocker.patch("grizzly.target.puppet_target.process_iter", autospec=True)
    mocker.patch("grizzly.target.puppet_target.sleep", autospec=True)
    fake_time = mocker.patch("grizzly.target.puppet_target.time", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    target = PuppetTarget(str(fake_file), None, 300, 25, 5000, str(fake_file), 10)
    fake_kill = mocker.patch("grizzly.target.puppet_target.kill", autospec=True)
    # not running
    target.rl_countdown = 1
    fake_ffp.return_value.get_pid.return_value = None
    target.dump_coverage()
    assert not fake_kill.call_count
    assert fake_ffp.return_value.get_pid.call_count == 1
    assert fake_proc_iter.call_count == 0
    # gcda not found
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.get_pid.return_value = 100
    fake_time.side_effect = (0, 1, 10)
    target.dump_coverage()
    assert fake_kill.call_count == 2
    assert fake_proc_iter.call_count == 2
    assert fake_ffp.return_value.is_healthy.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # browser crashes
    fake_ffp.return_value.is_healthy.side_effect = (True, False)
    fake_time.side_effect = None
    fake_time.return_value = 1.0
    target.dump_coverage()
    assert fake_kill.call_count == 2
    assert fake_proc_iter.call_count == 1
    assert fake_ffp.return_value.is_healthy.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # timeout while waiting for files
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.is_healthy.side_effect = None
    fake_ffp.return_value.get_pid.return_value = 100
    fake_proc_iter.return_value = (
        mocker.Mock(info={"pid": 100, "ppid": 0, "open_files": (mocker.Mock(path="a.gcda"),)}),
    )
    fake_time.side_effect = (0, 1, 20, 20)
    target.dump_coverage(timeout=15)
    assert fake_kill.call_count == 3
    assert fake_proc_iter.call_count == 2
    assert fake_ffp.return_value.is_healthy.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # wait for files (success)
    fake_ffp.return_value.get_pid.return_value = 100
    fake_time.side_effect = None
    fake_time.return_value = 1.0
    fake_proc_iter.side_effect = (
        (
            mocker.Mock(info={"pid": 100, "ppid": 0, "open_files": (mocker.Mock(path="a.bin"), mocker.Mock(path="/a/s/d"))}),
            mocker.Mock(info={"pid": 101, "ppid": 100, "open_files": None}),
            mocker.Mock(info={"pid": 999, "ppid": 0, "open_files": None})
        ),
        (
            mocker.Mock(info={"pid": 100, "ppid": 0, "open_files": (mocker.Mock(path="a.gcda"),)}),
        ),
        (
            mocker.Mock(info={"pid": 100, "ppid": 0, "open_files": (mocker.Mock(path="a.bin"),)}),
            mocker.Mock(info={"pid": 999, "ppid": 0, "open_files": (mocker.Mock(path="ignore.gcda"),)})
        )
    )
    target.dump_coverage()
    assert fake_proc_iter.call_count == 3
    assert fake_kill.call_count == 2

def test_puppet_target_05(mocker, tmp_path):
    """test is_idle()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.cpu_usage.return_value = [(999, 30), (998, 20), (997, 10)]
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), None, 300, 25, 5000, str(fake_file), 10) as target:
        assert not target.is_idle(0)
        assert not target.is_idle(25)
        assert target.is_idle(50)

def test_puppet_target_06(mocker, tmp_path):
    """test PuppetTarget.monitor"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), None, 300, 25, 5000, str(fake_file), 25) as target:
        fake_ffp.return_value.is_running.return_value = False
        fake_ffp.return_value.is_healthy.return_value = False
        assert target.monitor is not None
        assert not target.monitor.is_healthy()
        assert not target.monitor.is_running()
        fake_ffp.return_value.is_running.return_value = True
        fake_ffp.return_value.is_healthy.return_value = True
        assert target.monitor.is_healthy()
        assert target.monitor.is_running()
        fake_ffp.return_value.launches = 123
        assert target.monitor.launches == 123
        fake_ffp.return_value.log_length.return_value = 100
        assert target.monitor.log_length("stdout") == 100
        target.monitor.clone_log("somelog")
        assert fake_ffp.return_value.clone_log.call_count == 1

def test_puppet_target_07(mocker, tmp_path):
    """test PuppetTarget with GRZ_BROWSER_LOGS set"""
    browser_logs = (tmp_path / "browser_logs")
    fake_getenv = mocker.patch("grizzly.target.puppet_target.getenv", autospec=True)
    fake_getenv.return_value = str(browser_logs)
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), None, 300, 25, 5000, str(fake_file), 35) as target:
        target.launch("launch_target_page")
        assert target._browser_logs == str(browser_logs)
        assert browser_logs.is_dir()
        target.cleanup()
        assert target._browser_logs is None
    assert fake_ffp.return_value.save_logs.call_count == 1



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import abc
import hashlib
import json
import logging
import platform
import os
import re
import shutil
import tarfile
import tempfile
import time
import zipfile

import fasteners
import psutil

# import FuzzManager utilities
from Collector.Collector import Collector
from FTB.ProgramConfiguration import ProgramConfiguration
from FTB.Signatures.CrashInfo import CrashInfo

# check if boto is available for S3FuzzManager reporter
try:
    import boto3
    import botocore
    _boto_import_error = None  # pylint: disable=invalid-name
    logging.getLogger("botocore").setLevel(logging.WARNING)
    logging.getLogger("boto3").setLevel(logging.WARNING)
    logging.getLogger("s3transfer").setLevel(logging.WARNING)
except ImportError as err:
    _boto_import_error = err  # pylint: disable=invalid-name

from .stack_hasher import Stack
from .utils import grz_tmp

__all__ = ("FilesystemReporter", "FuzzManagerReporter", "Report", "S3FuzzManagerReporter")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

log = logging.getLogger("grizzly")  # pylint: disable=invalid-name


class Report(object):
    DEFAULT_MAJOR = "NO_STACK"
    DEFAULT_MINOR = "0"
    MAX_LOG_SIZE = 1048576  # 1MB

    __slots__ = ("_crash_info", "log_aux", "log_err", "log_out", "path", "prefix", "stack")

    def __init__(self, log_path, log_map, size_limit=MAX_LOG_SIZE):
        self._crash_info = None
        self.log_aux = log_map.get("aux") if log_map is not None else None
        self.log_err = log_map.get("stderr") if log_map is not None else None
        self.log_out = log_map.get("stdout") if log_map is not None else None
        self.path = log_path

        # tail logs if needed
        if size_limit < 1:
            log.warning("No limit set on report log size!")
        elif os.path.isdir(log_path):
            for fname in os.listdir(log_path):
                log_file_path = os.path.join(log_path, fname)
                if os.path.isfile(log_file_path):
                    Report.tail(log_file_path, size_limit)

        # look through logs one by one until we find a stack
        # NOTE: order matters aux->stderr->stdout
        for scan_log in (self.log_aux, self.log_err, self.log_out):
            if scan_log is None:
                continue
            with open(os.path.join(log_path, scan_log), "rb") as log_fp:
                stack = Stack.from_text(log_fp.read().decode("utf-8", errors="ignore"))
            if stack.frames:
                self.prefix = "%s_%s" % (stack.minor[:8], time.strftime("%Y-%m-%d_%H-%M-%S"))
                self.stack = stack
                break
        else:
            self.stack = None
            self.prefix = "%s_%s" % (self.DEFAULT_MINOR, time.strftime("%Y-%m-%d_%H-%M-%S"))

    def cleanup(self):
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)

    @staticmethod
    def crash_hash(crash_info):
        """Create CrashInfo object from logs.

        Args:
            crash_info (CrashInfo): Binary file being tested.

        Returns:
            str: Hash of the raw signature of the crash.
        """
        max_frames = Report.crash_signature_max_frames(crash_info, 5)
        sig = Report.crash_signature(crash_info, max_frames)
        return hashlib.sha1(sig.rawSignature.encode("utf-8")).hexdigest()[:16]

    def crash_info(self, target_binary):
        """Create CrashInfo object from logs.

        Args:
            target_binary (str): Binary file being tested.

        Returns:
            CrashInfo: CrashInfo based on Result log data.
        """
        if self._crash_info is None:
            # read in the log files and create a CrashInfo object
            aux_data = None
            if self.log_aux is not None:
                with open(os.path.join(self.path, self.log_aux), "rb") as log_fp:
                    aux_data = log_fp.read().decode("utf-8", errors="ignore").splitlines()
            stderr_file = os.path.join(self.path, self.log_err)
            stdout_file = os.path.join(self.path, self.log_out)
            # create ProgramConfiguration that can be reported to a FM server
            if os.path.isfile("%s.fuzzmanagerconf" % (target_binary,)):
                # attempt to use "<target_binary>.fuzzmanagerconf"
                fm_cfg = ProgramConfiguration.fromBinary(target_binary)
            else:
                log.debug("'%s.fuzzmanagerconf' does not exist", target_binary)
                fm_cfg = None
            if fm_cfg is None:
                log.debug("creating ProgramConfiguration")
                cpu = platform.machine().lower()
                fm_cfg = ProgramConfiguration(
                    os.path.basename(target_binary),
                    "x86_64" if cpu == "amd64" else cpu,
                    platform.system())
            with open(stderr_file, "rb") as err_fp, open(stdout_file, "rb") as out_fp:
                self._crash_info = CrashInfo.fromRawCrashData(
                    out_fp.read().decode("utf-8", errors="ignore").splitlines(),
                    err_fp.read().decode("utf-8", errors="ignore").splitlines(),
                    fm_cfg,
                    auxCrashData=aux_data)
        return self._crash_info

    @staticmethod
    def crash_signature(crash_info, max_frames=5):
        return crash_info.createCrashSignature(
            maxFrames=Report.crash_signature_max_frames(crash_info, max_frames))

    @staticmethod
    def crash_signature_max_frames(crash_info, suggested_frames=8):
        if set(crash_info.backtrace) & {
                "std::panicking::rust_panic",
                "std::panicking::rust_panic_with_hook",
        }:
            # rust panic adds 5-6 frames of noise at the top of the stack
            suggested_frames += 6
        return suggested_frames

    @classmethod
    def from_path(cls, path, size_limit=MAX_LOG_SIZE):
        """Create Report from a directory containing logs.

        Args:
            path (str): Directory containing log files.
            size_limit (int): Maximum size in bytes of a log file.

        Returns:
            Report: Result object based on log data.
        """
        return cls(path, Report.select_logs(path), size_limit=size_limit)

    @property
    def major(self):
        if self.stack and self.stack.major is not None:
            return self.stack.major
        return self.DEFAULT_MAJOR

    @property
    def minor(self):
        if self.stack and self.stack.minor is not None:
            return self.stack.minor
        return self.DEFAULT_MINOR

    @property
    def preferred(self):
        return self.log_aux if self.log_aux is not None else self.log_err

    @staticmethod
    def select_logs(log_path):
        if not os.path.isdir(log_path):
            raise IOError("log_path does not exist %r" % log_path)
        log_files = os.listdir(log_path)
        if not log_files:
            raise IOError("No logs found in %r" % log_path)
        logs = {"aux": None, "stderr": None, "stdout": None}

        # order by creation date because the oldest log is likely the cause of the issue
        log_files.sort(key=lambda x: os.stat(os.path.join(log_path, x)).st_mtime)

        # pattern to identify the ASan crash triggered when the parent process goes away
        re_e10s_forced = re.compile(r"""
            ==\d+==ERROR:.+?SEGV\son.+?0x[0]+\s\(.+?T2\).+?
            #0\s+0x[0-9a-f]+\sin\s+mozilla::ipc::MessageChannel::OnChannelErrorFromLink
            """, re.DOTALL | re.VERBOSE)

        # this is a list of *San error reports to prioritize
        # ASan reports not included below (deprioritized):
        # stack-overflow, BUS, failed to allocate, detected memory leaks
        interesting_sanitizer_tokens = (
            "use-after-", "-buffer-overflow on", ": SEGV on ", "access-violation on ",
            "negative-size-param", "attempting free on ", "-param-overlap")

        # look for sanitizer (ASan, UBSan, etc...) logs
        for fname in (log_file for log_file in log_files if "asan" in log_file):
            # grab first chunk of log to help triage
            with open(os.path.join(log_path, fname), "r") as log_fp:
                log_data = log_fp.read(4096)

            # look for interesting crash info in the log
            if "==ERROR:" in log_data:
                # check for e10s forced crash
                if re_e10s_forced.search(log_data) is not None:
                    continue
                # make sure there is something that looks like a stack frame in the log
                if "#0 " in log_data:
                    logs["aux"] = fname
                    if any(x in log_data for x in interesting_sanitizer_tokens):
                        break  # this is the likely cause of the crash
                    continue  # probably the most interesting but lets keep looking

            # UBSan error (non-ASan builds)
            if ": runtime error: " in log_data:
                logs["aux"] = fname

            # catch all (choose the one with info for now)
            if logs["aux"] is None and os.stat(os.path.join(log_path, fname)).st_size:
                logs["aux"] = fname

        # look for Valgrind logs
        if logs["aux"] is None:
            for fname in (log_file for log_file in log_files if "valgrind" in log_file):
                if os.stat(os.path.join(log_path, fname)).st_size:
                    logs["aux"] = fname
                    break

        # prefer ASan logs over minidump logs
        if logs["aux"] is None:
            re_dump_req = re.compile(r"\d+\|0\|.+?\|google_breakpad::ExceptionHandler::WriteMinidump")
            for fname in (log_file for log_file in log_files if "minidump" in log_file):
                with open(os.path.join(log_path, fname), "r") as log_fp:
                    log_data = log_fp.read(4096)
                    # this will select log that contains "Crash|SIGSEGV|" or
                    # the desired "DUMP_REQUESTED" log
                    # TODO: review this it may be too strict
                    # see https://searchfox.org/mozilla-central/source/accessible/ipc/DocAccessibleParent.cpp#452
                    if "Crash|DUMP_REQUESTED|" not in log_data or re_dump_req.search(log_data):
                        logs["aux"] = fname
                        break

        # look for ffpuppet worker logs, worker logs should be used if nothing else is available
        if logs["aux"] is None:
            for fname in (log_file for log_file in log_files if "ffp_worker" in log_file):
                if logs["aux"] is not None:
                    # we only expect one log here...
                    log.warning("aux log previously selected: %s, overwriting!", logs["aux"])
                logs["aux"] = fname

        for fname in log_files:
            if "stderr" in fname:
                logs["stderr"] = fname
            elif "stdout" in fname:
                logs["stdout"] = fname

        return logs

    @staticmethod
    def tail(in_file, size_limit):
        assert size_limit > 0
        if os.stat(in_file).st_size <= size_limit:
            return
        with open(in_file, "rb") as in_fp:
            in_fp.seek(0, os.SEEK_END)
            dump_pos = max((in_fp.tell() - size_limit), 0)
            in_fp.seek(dump_pos)
            out_fd, out_file = tempfile.mkstemp(
                prefix="taillog_",
                dir=grz_tmp())
            os.close(out_fd)
            with open(out_file, "wb") as out_fp:
                out_fp.write(b"[LOG TAILED]\n")
                shutil.copyfileobj(in_fp, out_fp, 0x10000)  # 64KB chunks
        os.remove(in_file)
        shutil.move(out_file, in_file)


class Reporter(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _process_report(self, report):
        pass

    @abc.abstractmethod
    def _reset(self):
        pass

    @abc.abstractmethod
    def _submit_report(self, report, test_cases):
        pass

    def submit(self, test_cases, log_path=None, report=None):
        """Submit report containing results. Either `log_path` or `report` must
        be specified.

        Args:
            test_cases (iterable): A collection of testcases, ordered newest to oldest,
                                   the newest being the mostly likely to trigger
                                   the result (crash, assert... etc).
            log_path (str): Path to logs from the Target. A Report will
                            be created from this.
            report (Report): Report to submit.

        Returns:
            None
        """
        if log_path is not None:
            assert report is None, "Only 'log_path' or 'report' can be specified!"
            if not os.path.isdir(log_path):
                raise IOError("No such directory %r" % log_path)
            report = Report.from_path(log_path)
        elif report is not None:
            assert isinstance(report, Report)
        else:
            raise AssertionError("Either 'log_path' or 'report' must be specified!")
        self._process_report(report)
        self._submit_report(report, test_cases)
        if report is not None:
            report.cleanup()
        self._reset()


class FilesystemReporter(Reporter):
    DISK_SPACE_ABORT = 512 * 1024 * 1024  # 512 MB

    def __init__(self, report_path=None, major_bucket=True):
        self.major_bucket = major_bucket
        self.report_path = os.path.join(os.getcwd(), "results") if report_path is None else report_path

    def _process_report(self, report):
        pass

    def _reset(self):
        pass

    def _submit_report(self, report, test_cases):
        # create major bucket directory in working directory if needed
        if self.major_bucket:
            dest_path = os.path.join(self.report_path, report.major[:16])
        else:
            dest_path = self.report_path
        if not os.path.isdir(dest_path):
            os.makedirs(dest_path)
        # dump test cases and the contained files to working directory
        for test_number, test_case in enumerate(test_cases):
            dump_path = os.path.join(dest_path, "%s-%d" % (report.prefix, test_number))
            if not os.path.isdir(dump_path):
                os.mkdir(dump_path)
            test_case.dump(dump_path, include_details=True)
        # move logs into bucket directory
        log_path = os.path.join(dest_path, "%s_%s" % (report.prefix, "logs"))
        if os.path.isdir(log_path):
            log.warning("Report log path exists %r", log_path)
        shutil.move(report.path, log_path)
        # avoid filling the disk
        free_space = psutil.disk_usage(log_path).free
        if free_space < self.DISK_SPACE_ABORT:
            raise RuntimeError("Running low on disk space (%0.1fMB)" % (free_space / 1048576.0,))


class FuzzManagerReporter(Reporter):
    FM_CONFIG = os.path.join(os.path.expanduser("~"), ".fuzzmanagerconf")
    # max number of times to report a non-frequent signature to FuzzManager
    MAX_REPORTS = 10

    # testcase quality values
    QUAL_REDUCED_RESULT = 0  # the final reduced testcase
    QUAL_REDUCED_ORIGINAL = 1  # the original used for successful reduction
    QUAL_REPRODUCIBLE = 4  # the testcase was reproducible
    QUAL_UNREDUCED = 5  # haven't attempted reduction yet (1st attempt, generic reducer)
    QUAL_REQUEST_SPECIFIC = 6  # platform specific reduction requested (2nd attempt)
    QUAL_NO_TESTCASE = 7  # no testcase was detected (could be the "testcase" is not a testcase)
    QUAL_REDUCER_BROKE = 8  # the testcase was reproducible, but broke during reduction
    QUAL_REDUCER_ERROR = 9  # reducer error
    QUAL_NOT_REPRODUCIBLE = 10  # could not reproduce the testcase

    def __init__(self, target_binary, tool=None):
        self._extra_metadata = {}
        self.force_report = False
        self.quality = self.QUAL_UNREDUCED
        self.target_binary = target_binary
        self.tool = tool  # optional tool name

    @staticmethod
    def create_crash_info(report, target_binary):
        # TODO: this is here to preserve the old way of operation (used by reducer)
        return report.crash_info(target_binary)

    def _reset(self):
        self._extra_metadata = {}

    @staticmethod
    def sanity_check(bin_file):
        """Perform FuzzManager sanity check.

        Args:
            bin_file (str): Binary file being tested.

        Returns:
            None
        """
        if not os.path.isfile(FuzzManagerReporter.FM_CONFIG):
            raise IOError("Missing: %s" % FuzzManagerReporter.FM_CONFIG)
        if not os.path.isfile("".join([bin_file, ".fuzzmanagerconf"])):
            raise IOError("Missing: %s.fuzzmanagerconf" % (bin_file,))
        ProgramConfiguration.fromBinary(bin_file)

    @classmethod
    def quality_name(cls, value):
        for name in dir(cls):
            if name.startswith("QUAL_") and getattr(cls, name) == value:
                return name
        return "unknown quality (%r)" % (value,)

    def _process_report(self, report):
        self._process_rr_trace(report)

    def _process_rr_trace(self, report):
        # don't report large files to FuzzManager
        trace_path = os.path.join(report.path, "rr-traces")
        if os.path.isdir(trace_path):
            log.info("Ignored rr trace")
            self._extra_metadata["rr-trace"] = "ignored"
            # remove traces so they are not uploaded to FM (because they are huge)
            # use S3FuzzManagerReporter instead
            shutil.rmtree(trace_path)

    @staticmethod
    def _ignored(report):
        # This is here to prevent reporting stack-less crashes
        # that were caused by system OOM or bogus other crashes
        log_file = os.path.join(report.path, report.preferred)
        with open(log_file, "rb") as log_fp:
            log_data = log_fp.read().decode("utf-8", errors="ignore")
        mem_errs = (
            "ERROR: Failed to mmap",
            ": AddressSanitizer failed to allocate")
        for msg in mem_errs:
            if msg in log_data and "#0 " not in log_data:
                return True
        if log_data.startswith("VEX temporary storage exhausted."):
            # ignore Valgrind crashes
            return True
        return False

    def _submit_report(self, report, test_cases):
        # prepare data for submission as CrashInfo
        crash_info = report.crash_info(self.target_binary)
        assert crash_info is not None

        # search for a cached signature match and if the signature
        # is already in the cache and marked as frequent, don't bother submitting
        with fasteners.process_lock.InterProcessLock(os.path.join(grz_tmp(), "fm_sigcache.lock")):
            collector = Collector()
            cache_sig_file, cache_metadata = collector.search(crash_info)
            if cache_metadata is not None:
                if cache_metadata["frequent"]:
                    log.info("Frequent crash matched existing signature: %s",
                             cache_metadata["shortDescription"])
                    if not self.force_report:
                        return
                elif "bug__id" in cache_metadata:
                    log.info("Crash matched existing signature (bug %s): %s",
                             cache_metadata["bug__id"],
                             cache_metadata["shortDescription"])
                    # we will still report this one, but no more
                    cache_metadata["frequent"] = True
                # there is already a signature, initialize count
                cache_metadata.setdefault("_grizzly_seen_count", 0)
            else:
                # there is no signature, create one locally so we can count
                # the number of times we've seen it
                max_frames = Report.crash_signature_max_frames(crash_info)
                cache_sig_file = collector.generate(crash_info, numFrames=max_frames)
                cache_metadata = {
                    "_grizzly_seen_count": 0,
                    "frequent": False,
                    "shortDescription": crash_info.createShortSignature()}
            if cache_sig_file is None:
                if self._ignored(report):
                    log.info("Report is unsupported and is in ignore list")
                    return
                log.warning("Report is unsupported by FM, saved to %r", report.path)
                # TODO: we should check if stackhasher failed too
                raise RuntimeError("Failed to create FM signature")
            # limit the number of times we report per cycle
            cache_metadata["_grizzly_seen_count"] += 1
            if cache_metadata["_grizzly_seen_count"] >= self.MAX_REPORTS:
                # we will still report this one, but no more
                cache_metadata["frequent"] = True
            metadata_file = cache_sig_file.replace(".signature", ".metadata")
            with open(metadata_file, "w") as meta_fp:
                json.dump(cache_metadata, meta_fp)

        # dump test cases and the contained files to working directory
        test_case_meta = []
        for test_number, test_case in enumerate(test_cases):
            test_case_meta.append([test_case.adapter_name, test_case.input_fname])
            dump_path = os.path.join(report.path, "%s-%d" % (report.prefix, test_number))
            if not os.path.isdir(dump_path):
                os.mkdir(dump_path)
            test_case.dump(dump_path, include_details=True)
        crash_info.configuration.addMetadata({"grizzly_input": repr(test_case_meta)})
        if test_cases:
            environ_string = " ".join("=".join(kv) for kv in test_cases[0].env_vars.items())
            crash_info.configuration.addMetadata({"recorded_envvars": environ_string})
        else:
            self.quality = self.QUAL_NO_TESTCASE
        crash_info.configuration.addMetadata(self._extra_metadata)

        # grab screen log
        if os.getenv("WINDOW") is not None:
            screen_log = ".".join(["screenlog", os.getenv("WINDOW")])
            if os.path.isfile(screen_log):
                target_log = os.path.join(report.path, "screenlog.txt")
                shutil.copyfile(screen_log, target_log)
                Report.tail(target_log, 10240)  # limit to last 10K

        # add results to a zip file
        zip_name = "%s.zip" % (report.prefix,)
        with zipfile.ZipFile(zip_name, mode="w", compression=zipfile.ZIP_DEFLATED) as zip_fp:
            # add test files
            for dir_name, _, dir_files in os.walk(report.path):
                arc_path = os.path.relpath(dir_name, report.path)
                for file_name in dir_files:
                    zip_fp.write(
                        os.path.join(dir_name, file_name),
                        arcname=os.path.join(arc_path, file_name))

        # override tool name if specified
        if self.tool is not None:
            collector.tool = self.tool

        # announce shortDescription if crash is not in a bucket
        if cache_metadata["_grizzly_seen_count"] == 1 and not cache_metadata["frequent"]:
            log.info("Submitting new crash %r", cache_metadata["shortDescription"])
        # submit results to the FuzzManager server
        new_entry = collector.submit(crash_info, testCase=zip_name, testCaseQuality=self.quality)
        log.info("Logged %d with quality %d", new_entry["id"], self.quality)

        # remove zipfile
        if os.path.isfile(zip_name):
            os.remove(zip_name)


class S3FuzzManagerReporter(FuzzManagerReporter):
    @staticmethod
    def compress_rr_trace(src, dest):
        # resolve symlink to latest trace available
        latest_trace = os.path.realpath(os.path.join(src, "latest-trace"))
        assert os.path.isdir(latest_trace), "missing latest-trace directory"
        rr_arc = os.path.join(dest, "rr.tar.bz2")
        log.debug("creating %r from %r", rr_arc, latest_trace)
        with tarfile.open(rr_arc, "w:bz2") as arc_fp:
            arc_fp.add(latest_trace, arcname=os.path.basename(latest_trace))
        # remove path containing uncompressed traces
        shutil.rmtree(src)
        return rr_arc

    def _process_report(self, report):
        self._process_rr_trace(report)

    def _process_rr_trace(self, report):
        trace_path = os.path.join(report.path, "rr-traces")
        if not os.path.isdir(trace_path):
            return None
        s3_bucket = os.getenv("GRZ_S3_BUCKET")
        assert s3_bucket is not None
        # check for existing minor hash in S3
        s3 = boto3.resource("s3")
        s3_key = "rr-%s.tar.bz2" % (report.minor,)
        s3_url = "http://%s.s3.amazonaws.com/%s" % (s3_bucket, s3_key)
        try:
            s3.Object(s3_bucket, s3_key).load()  # HEAD, doesn't fetch the whole object
        except botocore.exceptions.ClientError as exc:
            if exc.response["Error"]["Code"] == "404":
                # The object does not exist.
                pass
            else:
                # Something else has gone wrong.
                raise
        else:
            # The object already exists.
            log.info("RR trace exists at %s", s3_url)
            self._extra_metadata["rr-trace"] = s3_url
            # remove traces so they are not reported to FM
            shutil.rmtree(trace_path)
            return s3_url

        # Upload to S3
        rr_arc = self.compress_rr_trace(trace_path, report.path)
        s3.meta.client.upload_file(rr_arc, s3_bucket, s3_key, ExtraArgs={"ACL": "public-read"})
        os.unlink(rr_arc)
        self._extra_metadata["rr-trace"] = s3_url
        return s3_url

    @staticmethod
    def sanity_check(bin_file):
        if os.getenv("GRZ_S3_BUCKET") is None:
            raise EnvironmentError("'GRZ_S3_BUCKET' is not set in environment")
        if _boto_import_error is not None:
            raise _boto_import_error  # pylint: disable=raising-bad-type
        FuzzManagerReporter.sanity_check(bin_file)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Manage Grizzly status reports."""
from json import dump, load
from logging import getLogger
import os
from tempfile import mkstemp
from time import time

import fasteners

from .utils import grz_tmp

__all__ = ("ReducerStats", "Status")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger("status")


class Status(object):
    """Status holds status information for the Grizzly session.
    There can be multiple readers of the data but only a single writer.
    """
    AGE_LIMIT = 3600  # 1 hour
    PATH = grz_tmp("status")
    REPORT_FREQ = 60

    __slots__ = (
        "_lock", "data_file", "ignored", "iteration", "log_size", "results",
        "start_time", "test_name", "timestamp")

    def __init__(self, data_file, start_time):
        assert ".json" in data_file
        assert isinstance(start_time, float)
        self._lock = fasteners.process_lock.InterProcessLock("%s.lock" % (data_file,))
        self.data_file = data_file
        self.ignored = 0
        self.iteration = 0
        self.log_size = 0
        self.results = 0
        self.start_time = start_time
        self.test_name = None
        self.timestamp = start_time

    def cleanup(self):
        """Remove data file.

        Args:
            None

        Returns:
            None
        """
        if self.data_file is None:
            return
        with self._lock:
            try:
                if os.path.isfile(self.data_file):
                    os.unlink(self.data_file)
            except OSError:  # pragma: no cover
                LOG.warning("Failed to delete %r", self.data_file)
            lock_file = "%s.lock" % (self.data_file,)
            self.data_file = None
        try:
            os.unlink(lock_file)
        except OSError:  # pragma: no cover
            pass

    @property
    def duration(self):
        """Calculate the number of second since start() was called

        Args:
            None

        Returns:
            int: Total runtime in seconds since start() was called
        """
        return max(self.timestamp - self.start_time, 0)

    @classmethod
    def load(cls, data_file):
        """Read Grizzly status report.

        Args:
            data_file (str): JSON file that contains status data.

        Returns:
            Status: Loaded status object or None
        """
        data = None
        lock_file = "%s.lock" % (data_file,)
        lock = fasteners.process_lock.InterProcessLock(lock_file)
        # there is race between looking up status files and loading them
        # if a status item is removed before it can be loaded a lock file
        # would be left behind from this load attempt
        cleanup = not lock.exists()
        with lock:
            try:
                with open(data_file, "r") as out_fp:
                    data = load(out_fp)
            except IOError:
                LOG.debug("%r does not exist", data_file)
            except ValueError:
                LOG.debug("failed to load json")
        if cleanup:
            # the lock file should be removed if it was created here
            try:
                os.unlink(lock_file)
            except OSError:  # pragma: no cover
                pass
        if data is None:
            return None
        if "start_time" not in data:
            LOG.debug("invalid status json file")
            return None
        status = cls(data_file, data["start_time"])
        for attr, value in data.items():
            setattr(status, attr, value)
        return status

    @classmethod
    def loadall(cls):
        """Read all Grizzly status reports found in cls.PATH.

        Args:
            None

        Returns:
            Generator: Status objects stored in cls.PATH.
        """
        if not os.path.isdir(cls.PATH):
            return
        for data_file in os.listdir(cls.PATH):
            if not data_file.endswith(".json"):
                continue
            status = cls.load(os.path.join(cls.PATH, data_file))
            if status is None:
                continue
            yield status

    @property
    def rate(self):
        """Calculate the number of iterations performed per second since start() was called

        Args:
            None

        Returns:
            float: Number of iterations performed per second
        """
        return self.iteration / float(self.duration) if self.duration > 0 else 0

    @property
    def _data(self):
        return {
            "ignored": self.ignored,
            "iteration": self.iteration,
            "log_size": self.log_size,
            "results": self.results,
            "start_time": self.start_time,
            "test_name": self.test_name,
            "timestamp": self.timestamp}

    def report(self, force=False, report_freq=REPORT_FREQ):
        """Write Grizzly status report. Reports are only written when the duration
        of time since the previous report was created exceeds `report_freq` seconds

        Args:
            force (bool): Ignore report frequently limiting.
            report_freq (int): Minimum number of seconds between writes.

        Returns:
            bool: Returns true if the report was successful otherwise false
        """
        now = time()
        if not force and now < (self.timestamp + report_freq):
            return False
        self.timestamp = now
        with self._lock:
            with open(self.data_file, "w") as out_fp:
                dump(self._data, out_fp)
        return True

    @classmethod
    def start(cls):
        """Create a unique Status object.

        Args:
            None

        Returns:
            Status: Ready to be used to report Grizzly status
        """
        if not os.path.isdir(cls.PATH):
            try:
                os.mkdir(cls.PATH)
            except OSError:  # pragma: no cover
                if not os.path.isdir(cls.PATH):
                    raise
        tfd, filepath = mkstemp(dir=cls.PATH, prefix="grzstatus_", suffix=".json")
        os.close(tfd)
        status = cls(filepath, time())
        status.report(force=True)
        return status


class ReducerStats(object):
    """ReducerStats holds stats for the Grizzly reducer.
    """
    FILE = "reducer-stats.json"
    PATH = grz_tmp("status")

    __slots__ = ("_file", "_lock", "error", "failed", "passed")

    def __init__(self):
        self._file = os.path.join(self.PATH, self.FILE)
        self._lock = None
        self.error = 0
        self.failed = 0
        self.passed = 0

    def __enter__(self):
        self._lock = fasteners.process_lock.InterProcessLock("%s.lock" % (self._file,))
        self._lock.acquire()
        try:
            with open(self._file, "r") as in_fp:
                data = load(in_fp)
            self.error = data["error"]
            self.failed = data["failed"]
            self.passed = data["passed"]
        except KeyError:
            LOG.debug("invalid status data in %r", self._file)
        except IOError:
            LOG.debug("%r does not exist", self._file)
        except ValueError:
            LOG.debug("failed to load stats from %r", self._file)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            with open(self._file, "w") as out_fp:
                dump({
                    "error": self.error,
                    "failed": self.failed,
                    "passed": self.passed}, out_fp)
        finally:
            if self._lock:
                self._lock.release()
                self._lock = None



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

__author__ = "Jesse Schwartzentruber"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


class ReducerError(Exception):
    pass


class TestcaseError(ReducerError):
    pass


class NoTestcaseError(TestcaseError):
    pass


class CorruptTestcaseError(TestcaseError):
    pass



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from os.path import isdir, isfile

from .reduce import ReductionJob
from .strategies import strategies_by_name
from ..args import CommonArgs


class ReducerArgs(CommonArgs):

    def __init__(self):
        super(ReducerArgs, self).__init__()
        self.parser.add_argument(
            "input",
            help="Test case or directory containing test cases")

        replay_args = self.parser.add_argument_group("Reduce Arguments")
        replay_args.add_argument(
            "--any-crash", action="store_true",
            help="Any crash is interesting, not only crashes which match the original first crash")
        replay_args.add_argument(
            "--environ",
            help="DEPRICATED: File containing line separated environment variables (VAR=value)" \
                "to be set in the firefox process.")
        replay_args.add_argument(
            "--idle-threshold", type=int, default=25,
            help="CPU usage threshold to mark the process as idle (default: %(default)s)")
        replay_args.add_argument(
            "--idle-timeout", type=int, default=60,
            help="Number of seconds to wait before polling testcase for idle (default: %(default)s)")
        replay_args.add_argument(
            "--min-crashes", type=int, default=1,
            help="Require the testcase to crash n times before accepting the result. (default: %(default)sx)")
        replay_args.add_argument(
            "--no-analysis", action="store_true",
            help="Disable analysis to auto-set --repeat/--min-crashes.")
        replay_args.add_argument(
            "--no-cache", action="store_true",
            help="Disable testcase caching")
        replay_args.add_argument(
            "--no-harness", action="store_true",
            help="Don't use the harness for sapphire redirection")
        replay_args.add_argument(
            "--reduce-file",
            help="Value passed to lithium's --testcase option, needed for testcase cache " \
                "(default: input param)")
        replay_args.add_argument(
            "--repeat", type=int, default=1,
            help="Try to run the testcase multiple times, for intermittent testcases (default: %(default)sx)")
        replay_args.add_argument(
            "--sig",
            help="Signature (JSON) file to match.")
        replay_args.add_argument(
            "--skip", type=int, default=0,
            help="Return interesting = False for the first n reductions (default: %(default)s)")
        replay_args.add_argument(
            "--static-timeout", action="store_true", dest="fixed_timeout",
            help="Disable automatically updating the iteration timeout.")
        replay_args.add_argument(
            "--strategy", nargs="+", default=list(), metavar="STRATEGY", dest="strategies",
            help="One or more strategies (space-separated). Available: %s (default: %s)"
            % (" ".join(sorted(strategies_by_name())), " ".join(ReductionJob.DEFAULT_STRATEGIES)))

        self.parser.epilog = "For addition help check out the wiki:" \
            " https://github.com/MozillaSecurity/grizzly/wiki"

    def sanity_check(self, args):
        super(ReducerArgs, self).sanity_check(args)

        if "input" not in self._sanity_skip:
            if not (isdir(args.input)
                    or (isfile(args.input) and (args.input.lower().endswith(".zip")
                                                or args.input.lower().endswith(".html")))):
                self.parser.error("Testcase should be a folder, zip, or html file")

        if args.sig is not None and not isfile(args.sig):
            self.parser.error("file not found: %r" % args.sig)

        if args.repeat < 1:
            self.parser.error("'--repeat' value must be positive")

        if args.min_crashes < 1:
            self.parser.error("'--min-crashes' value must be positive")

        if args.environ is not None and not isfile(args.environ):
            self.parser.error("file not found: %r" % args.environ)

        if args.strategies:
            known_strategies = set(strategies_by_name())
            for strategy in args.strategies:
                if strategy not in known_strategies:
                    self.parser.error("invalid strategy: %s" % (strategy,))
        else:
            args.strategies = None

        if args.reduce_file is None:
            args.reduce_file = args.input


class ReducerFuzzManagerIDArgs(ReducerArgs):

    def __init__(self):
        super(ReducerFuzzManagerIDArgs, self).__init__()

        # madhax alert!
        #
        # We need to modify the meaning of the 'input' positional to accept an int ID instead of a
        # local testcase. This is not possible with the public argparse API.
        #
        # refs: https://stackoverflow.com/questions/32807319/disable-remove-argument-in-argparse
        #       https://bugs.python.org/issue19462

        # look up the action for the positional `input` arg
        action = None
        for arg in self.parser._actions:
            if arg.dest == "input" and not arg.option_strings:
                action = arg
                break
        assert action is not None

        # modify it's type and help string
        action.type = int
        action.help = "FuzzManager ID to reduce"

        # ... and Bob's your uncle
        self._sanity_skip.add("input")


class ReducerFuzzManagerIDQualityArgs(ReducerFuzzManagerIDArgs):

    def __init__(self):
        super(ReducerFuzzManagerIDQualityArgs, self).__init__()
        self.parser.add_argument("--quality", type=int,
                                 help="Only try crashes with a given quality value")

    def sanity_check(self, args):
        super(ReducerFuzzManagerIDQualityArgs, self).sanity_check(args)

        if args.quality is not None and args.quality < 0:
            self.parser.error("'--quality' value must be positive or zero")



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import abc
import os


__all__ = ("Adapter", "AdapterError")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class AdapterError(Exception):
    """The base class for exceptions raised by an Adapter"""


class Adapter(metaclass=abc.ABCMeta):
    HARNESS_FILE = os.path.join(os.path.dirname(__file__), "harness.html")
    IGNORE_UNSERVED = True  # Only report test cases with served content
    NAME = None  # must be a unique string
    RELAUNCH = 0  # maximum iterations between Target relaunches (<1 use default)
    TEST_DURATION = 30  # maximum execution time per test

    __slots__ = ("_harness", "fuzz", "monitor", "remaining")

    #############################
    # Built-ins do NOT overload!
    #############################

    def __init__(self):
        if not isinstance(self.NAME, str):
            raise AdapterError("%s.NAME must be a string" % (type(self).__name__,))
        self._harness = None
        self.fuzz = dict()
        self.monitor = None
        # remaining can be used to indicate the number of TestCases remaining to process
        self.remaining = None

    def cleanup(self):
        """Automatically called once at shutdown.

        Args:
            None

        Returns:
            None
        """
        self.shutdown()

    def enable_harness(self, file_path=None):
        """Enable use of a harness during fuzzing. By default no harness is used.

        Args:
            file_path (str): Path to file to use as a harness. If None the default harness is used.

        Returns:
            None
        """
        if file_path is None:
            file_path = self.HARNESS_FILE
        with open(file_path, "rb") as in_fp:
            self._harness = in_fp.read()

    def get_harness(self):
        """Get the harness. Used internally by Grizzly.

        Args:
            None

        Returns:
            TestFile: The current harness
        """
        return self._harness

    @staticmethod
    def scan_path(path, ignore=("desktop.ini", "thumbs.db"), recursive=False):
        """Scan a path and yield the files within it.

        Args:
            path (str): Path to file or directory.
            ignore (iterable): Filenames to ignore.
            recursive (bool): Scan recursively into directories.

        Yields:
            str: Absolute path to files.
        """
        full_path = os.path.abspath(path)
        if os.path.isdir(full_path):
            for root, _, files in os.walk(full_path):
                for fname in files:
                    if fname in ignore or fname.startswith("."):
                        # skip ignored and hidden system files
                        continue
                    yield os.path.join(root, fname)
                if not recursive:
                    break
        elif os.path.isfile(full_path):
            yield full_path

    #############################
    # Methods to overload
    #############################

    @abc.abstractmethod
    def generate(self, testcase, server_map):
        """Automatically called. Populate testcase here.

        Args:
            testcase (TestCase): TestCase intended to be populated
            server_map (ServerMap): A ServerMap

        Returns:
            None
        """

    def on_served(self, testcase, served):
        """Optional. Automatically called after a test case is successfully served.

        Args:
            testcase (TestCase): TestCase that was served
            served (list): A list of file names served from testcase

        Returns:
            None
        """

    def on_timeout(self, testcase, served):
        """Optional. Automatically called if timeout occurs attempting to serve a test case.

        Args:
            testcase (TestCase): TestCase that was served
            served (list): A list of file names served from testcase

        Returns:
            None
        """
        self.on_served(testcase, served)

    def pre_launch(self):
        """Optional. Automatically called before launching the Target.

        Args:
            None

        Returns:
            None
        """

    def setup(self, input_path, server_map):
        """Optional. Automatically called once at startup.

        Args:
            input_path (str): Points to a file or directory passed by the user.
                              None is passed by default.
            server_map (ServerMap): A ServerMap

        Returns:
            None
        """

    def shutdown(self):
        """Optional. Automatically called once at shutdown.

        Args:
            None

        Returns:
            None
        """



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import basicConfig, DEBUG
from os import getenv
from sys import exit as sysexit

from .adapters import load
from .args import GrizzlyArgs
from .main import main

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


# TODO: This can go away once Adapters are loaded using
# setuptools entrypoints. It is only needed to get log output from
# load() because it is called before parse arguments (which
# is where basicConfig should be called).
if getenv("DEBUG"):
    basicConfig(
        format="%(levelname).1s %(name)s [%(asctime)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=DEBUG)
# load Adapters
load()
sysexit(main(GrizzlyArgs().parse_args()))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import logging
import os
import re
import sys
import tempfile

from Collector.Collector import Collector

from .args import ReducerFuzzManagerIDArgs
from .reduce import ReductionJob
from ..common import FuzzManagerReporter


LOG = logging.getLogger("grizzly.reduce.crash")


def crashentry_data(crash_id, raw=False):
    """Get the CrashEntry data for the specified FuzzManager crash

    Args:
        crash_id (int): ID of the requested crash on the server side
        raw (bool): include rawCrashData, rawStderr, rawStdout in result

    Returns:
        dict: crash entry data (crashmanager.models.CrashEntry)
    """
    coll = Collector()

    LOG.debug("crash %d, downloading metadata...", crash_id)

    url = "%s://%s:%d/crashmanager/rest/crashes/%s/" \
        % (coll.serverProtocol, coll.serverHost, coll.serverPort, crash_id)

    return coll.get(url, params={"include_raw": "1" if raw else "0"}).json()


def download_crash(crash_id):
    """Download testcase for the specified FuzzManager crash.

    Args:
        crash_id (int): ID of the requested crash on the server side

    Returns:
        str: Temporary filename of the testcase. Caller must remove when finished.
    """
    coll = Collector()

    LOG.debug("crash %d, downloading testcase...", crash_id)

    url = "%s://%s:%d/crashmanager/rest/crashes/%s/download/" \
        % (coll.serverProtocol, coll.serverHost, coll.serverPort, crash_id)

    response = coll.get(url)

    disp_m = re.match(r'^attachment; filename="(.*)"$',
                      response.headers.get("content-disposition", ""))

    if disp_m is None:
        raise RuntimeError("Server sent malformed response: %r" % (response,))

    prefix = "crash.%d." % (crash_id,)
    suffix = os.path.splitext(disp_m.group(1))[1]
    testcase_fd, testcase_fn = tempfile.mkstemp(prefix=prefix, suffix=suffix)
    with os.fdopen(testcase_fd, "wb") as testcase_fp:
        testcase_fp.write(response.content)

    return testcase_fn


def change_quality(crash_id, quality):
    """Update a FuzzManager crash entry quality.

    Args:
        crash_id (int): Crash ID on FuzzManager server
        quality (int): Quality constant defined in FuzzManagerReporter.QUAL_*

    Returns:
        None
    """
    LOG.info("Updating crash %d to quality %s", crash_id, FuzzManagerReporter.quality_name(quality))
    coll = Collector()

    url = "%s://%s:%d/crashmanager/rest/crashes/%d/" \
        % (coll.serverProtocol, coll.serverHost, coll.serverPort, crash_id)
    try:
        Collector().patch(url, data={"testcase_quality": quality})
    except RuntimeError as exc:
        # let 404's go .. evidently the crash was deleted
        if str(exc) == "Unexpected HTTP response: 404":
            LOG.warning("Failed to update (404), does the crash still exist?")
        else:
            raise


class CrashReductionJob(ReductionJob):
    __slots__ = ['_crash_id', '_fm_reporter', '_quality', '_testcase_path', '_tool_override',
                 '_was_interesting']

    def __init__(self, *args, **kwds):
        super(CrashReductionJob, self).__init__(*args, **kwds)
        self._crash_id = None
        self._fm_reporter = False
        self._quality = None
        self._testcase_path = None
        self._tool_override = False
        self._was_interesting = False

    def on_result(self, result_code):
        # only update quality of the original crash if we are reporting to FuzzManager
        if not self._fm_reporter:
            return

        if result_code == FuzzManagerReporter.QUAL_REDUCED_ORIGINAL:
            # reduce succeeded
            change_quality(self._crash_id, result_code)

        elif result_code == FuzzManagerReporter.QUAL_NOT_REPRODUCIBLE:
            if self._quality == FuzzManagerReporter.QUAL_UNREDUCED:
                # override result to request platform specific reduction
                result_code = FuzzManagerReporter.QUAL_REQUEST_SPECIFIC
            change_quality(self._crash_id, result_code)

        # for these cases, something went wrong. a reduce log/result would be really valuable
        elif result_code in {FuzzManagerReporter.QUAL_REDUCER_BROKE,
                             FuzzManagerReporter.QUAL_REDUCER_ERROR}:
            # for now just change the quality
            change_quality(self._crash_id, result_code)

        else:
            LOG.error("Got unhandled quality: %s", FuzzManagerReporter.quality_name(result_code))

    def on_interesting_crash(self, *args, **kwds):
        super(CrashReductionJob, self).on_interesting_crash(*args, **kwds)
        if self._was_interesting:
            return
        LOG.info("Crash %d reproduced!", self._crash_id)
        if self._fm_reporter:
            change_quality(self._crash_id, FuzzManagerReporter.QUAL_REPRODUCIBLE)
        self._was_interesting = True

    def run(self, *args, **kwds):
        try:
            return super(CrashReductionJob, self).run(*args, **kwds)
        finally:
            os.unlink(self._testcase_path)

    @classmethod
    def from_args(cls, args, target, status):
        LOG.info("Trying crash %d", args.input)

        try:
            crash_id = args.input
            testcase = download_crash(crash_id)
            tool_override = args.tool is None
            crash = crashentry_data(crash_id)
            quality = crash["testcase_quality"]
            if tool_override:
                args.tool = crash["tool"]
                LOG.info("Using toolname from crash: %s", args.tool)

            # reduce.main expects input to be a path to testcase
            args.input = testcase

            job = super(CrashReductionJob, cls).from_args(args, target, status)
            job._fm_reporter = args.fuzzmanager
            job._crash_id = crash_id
            job._tool_override = tool_override
            job._quality = quality
            job._testcase_path = testcase
            return job

        except:  # noqa
            os.unlink(testcase)
            raise


if __name__ == "__main__":
    sys.exit(CrashReductionJob.main(ReducerFuzzManagerIDArgs().parse_args()))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access

import pytest

from .server_map import InvalidURLError, MapCollisionError, Resource, ServerMap


def test_servermap_01():
    """test empty ServerMap"""
    srv_map = ServerMap()
    assert not srv_map.dynamic
    assert not srv_map.include
    assert not srv_map.redirect

def test_servermap_02(tmp_path):
    """test ServerMap dynamic responses"""
    def fake_cb():
        pass
    srv_map = ServerMap()
    srv_map.set_dynamic_response("url_01", fake_cb, mime_type="test/type")
    assert len(srv_map.dynamic) == 1
    assert "url_01" in srv_map.dynamic
    assert srv_map.dynamic["url_01"].mime == "test/type"
    assert callable(srv_map.dynamic["url_01"].target)
    assert srv_map.dynamic["url_01"].type == Resource.URL_DYNAMIC
    srv_map.set_dynamic_response("url_02", fake_cb, mime_type="foo")
    assert len(srv_map.dynamic) == 2
    assert not srv_map.include
    assert not srv_map.redirect
    with pytest.raises(MapCollisionError):
        srv_map.set_include("url_01", str(tmp_path))
    with pytest.raises(MapCollisionError):
        srv_map.set_redirect("url_01", "test_file")

def test_servermap_03(tmp_path):
    """test ServerMap includes"""
    srv_map = ServerMap()
    with pytest.raises(IOError, match="Include path not found: no_dir"):
        srv_map.set_include("test_url", "no_dir")
    assert not srv_map.include
    srv_map.set_include("url_01", str(tmp_path))
    assert len(srv_map.include) == 1
    assert "url_01" in srv_map.include
    assert srv_map.include["url_01"].target == str(tmp_path)
    # overwrite existing
    inc1 = (tmp_path / "includes" / "a")
    inc1.mkdir(parents=True)
    srv_map.set_include("url_01", str(inc1))
    assert srv_map.include["url_01"].target == str(inc1)
    # add another
    inc2 = (tmp_path / "includes" / "b")
    inc2.mkdir()
    srv_map.set_include("url_02", str(inc2))
    assert len(srv_map.include) == 2
    assert not srv_map.dynamic
    assert not srv_map.redirect
    with pytest.raises(MapCollisionError, match="URL collision on 'url_01'"):
        srv_map.set_redirect("url_01", "test_file")
    with pytest.raises(MapCollisionError):
        srv_map.set_dynamic_response("url_01", lambda: 0, mime_type="test/type")
    # test overlapping includes
    with pytest.raises(MapCollisionError, match=r"'url_01' and '\w+' include"):
        srv_map.set_include("url_01", str(tmp_path))
    inc3 = (tmp_path / "includes" / "b" / "c")
    inc3.mkdir()
    with pytest.raises(MapCollisionError, match=r"'url_01' and '\w+' include"):
        srv_map.set_include("url_01", str(inc3))

def test_servermap_04(tmp_path):
    """test ServerMap redirects"""
    srv_map = ServerMap()
    srv_map.set_redirect("url_01", "test_file", required=True)
    assert len(srv_map.redirect) == 1
    assert "url_01" in srv_map.redirect
    assert srv_map.redirect["url_01"].target == "test_file"
    assert srv_map.redirect["url_01"].required
    srv_map.set_redirect("url_02", "test_file", required=False)
    assert len(srv_map.redirect) == 2
    assert not srv_map.redirect["url_02"].required
    assert not srv_map.dynamic
    assert not srv_map.include
    with pytest.raises(MapCollisionError):
        srv_map.set_include("url_01", str(tmp_path))
    with pytest.raises(MapCollisionError):
        srv_map.set_dynamic_response("url_01", lambda: 0, mime_type="test/type")

def test_servermap_05():
    """test ServerMap._check_url()"""
    assert ServerMap._check_url("test") == "test"
    assert ServerMap._check_url("") == ""
    # only alpha-numeric is allowed
    with pytest.raises(InvalidURLError):
        ServerMap._check_url("asd!@#")
    # '..' should not be accepted
    with pytest.raises(InvalidURLError):
        ServerMap._check_url("/..")
    # cannot map more than one '/' deep
    with pytest.raises(InvalidURLError):
        ServerMap._check_url("/test/test")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
import pytest

from .target import sanitizer_opts, Target, TargetError
from .target_monitor import TargetMonitor

class SimpleTarget(Target):
    def cleanup(self):
        pass
    def close(self):
        pass
    @property
    def closed(self):
        pass
    def detect_failure(self, ignored, was_timeout):
        pass
    def launch(self, location, env_mod=None):
        pass
    @property
    def monitor(self):
        return self._monitor
    def save_logs(self, *args, **kwargs):
        pass

def test_target_01(tmp_path):
    """test creating a simple Target"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with pytest.raises(TargetError):
        SimpleTarget(str(fake_file), None, 321, 2, 3, "no_prefs", 25)
    target = SimpleTarget(str(fake_file), str(fake_file), 321, 2, 3, str(fake_file), 25)
    assert target.binary == str(fake_file)
    assert target.extension == str(fake_file)
    assert target.forced_close
    assert not target.is_idle(0)
    assert target.launch_timeout == 321
    assert target.log_size() == 0
    assert target.log_limit == 2
    assert target.memory_limit == 3
    assert target.rl_countdown == 0
    assert target.rl_reset == 25
    assert target.prefs == str(fake_file)
    assert not target.expect_close
    # test stubs
    target.add_abort_token("none!")
    target.dump_coverage()
    target.reverse(1, 2)

def test_target_02(mocker, tmp_path):
    """test setting Target.forced_close"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    getenv = mocker.patch("grizzly.target.target.getenv", autospec=True, return_value="0")
    with SimpleTarget(str(fake_file), None, 300, 25, 5000, None, 25) as target:
        assert not target.forced_close
        assert target.extension is None
        assert target.prefs is None
        target.rl_countdown = 1
        assert not target.expect_close
        target.rl_countdown = 0
        assert target.expect_close
    getenv.assert_called_with("GRZ_FORCED_CLOSE")

def test_target_03(mocker, tmp_path):
    """test Target.check_relaunch() and Target.step()"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with SimpleTarget(str(fake_file), None, 300, 25, 5000, None, 1) as target:
        target._monitor = mocker.Mock(spec=TargetMonitor)
        target._monitor.is_healthy.return_value = True
        # test skipping relaunch
        target.rl_countdown = 2
        target.step()
        assert target.rl_countdown == 1
        target.check_relaunch(wait=60)
        # test triggering relaunch
        target.rl_countdown = 1
        target.step()
        assert target.rl_countdown == 0
        target.check_relaunch(wait=0)
        # test with "crashed" process
        target._monitor.is_healthy.side_effect = (True, False)
        target.rl_countdown = 0
        target.step()
        mocker.patch("grizzly.target.target.sleep", autospec=True)
        target.check_relaunch(wait=5)

def test_sanitizer_opts_01(tmp_path):
    """test sanitizer_opts()"""
    # test empty string
    assert not sanitizer_opts("")
    # test single value
    opts = sanitizer_opts("test_value=true")
    assert len(opts) == 1
    assert opts["test_value"] == "true"
    # test multiple values
    opts = sanitizer_opts("a=1:b=-2:C=3")
    assert len(opts) == 3
    assert opts["a"] == "1"
    assert opts["b"] == "-2"
    assert opts["C"] == "3"
    # path parsing
    opts = sanitizer_opts("p1='z:/a':p2='x:\\a.1':p3='/test/path/':p4='':p5=\"x:/a.a\"")
    assert opts["p1"] == "'z:/a'"
    assert opts["p2"] == "'x:\\a.1'"
    assert opts["p3"] == "'/test/path/'"
    assert opts["p4"] == "''"
    assert opts["p5"] == "\"x:/a.a\""
    # platform specific parsing
    fake_file = tmp_path / "fake.log"
    opts = sanitizer_opts("bar=1:file='%s':foo=2" % (str(fake_file),))
    assert len(opts) == 3
    assert opts["file"] == "'%s'" % (str(fake_file),)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os

from .target_monitor import TargetMonitor

def test_target_monitor_01(tmp_path):
    """test a basic TargetMonitor"""
    class _BasicMonitor(TargetMonitor):
        # pylint: disable=no-self-argument
        def clone_log(_, log_id, offset=0):
            log_file = tmp_path / "test_log.txt"
            log_file.write_bytes(b"test")
            return str(log_file)
        def is_healthy(_):
            return True
        def is_running(_):
            return True
        @property
        def launches(_):
            return 1
        def log_length(_, log_id):
            return 100
    mon = _BasicMonitor()
    test_log = mon.clone_log("test_log", offset=0)
    assert os.path.isfile(test_log)
    assert mon.is_healthy()
    assert mon.is_running()
    assert mon.launches == 1
    assert mon.log_data("test_log") == b"test"
    assert mon.log_length("test_log") == 100



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from grizzly.common import Adapter

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class NoOpAdapter(Adapter):
    """
    This is an simple adapter that has very little overhead. It can be used
    to get a baseline iteration rate for Grizzly.
    """
    NAME = "no-op"

    def setup(self, *_):
        self.enable_harness()
        self.fuzz["test"] = "<!DOCTYPE html>\n" \
            "<html>\n" \
            "<head>\n" \
            "<script>window.close()</script>\n" \
            "</head>\n" \
            "</html>"

    def generate(self, testcase, _):
        testcase.add_from_data(self.fuzz["test"], testcase.landing_page)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import unicode_literals
import os
from grizzly.target.target import Target
from grizzly.common import Reporter
from grizzly.reduce import crash, reduce


class BaseFakeReporter(Reporter):
    def _process_report(self, _):
        pass

    def _reset(self):
        pass

    def _submit_report(self, *_args, **_kwds):
        pass


class FakeReduceStatus(object):
    "Stub to fake parts of ReduceStatus"

    def __init__(self):
        self.uid = 0
        self.duration = 0
        self.ignored = 0
        self.iteration = 0
        self.rate = 0
        self.results = 0
        self.reduce_error = 0
        self.reduce_fail = 0
        self.reduce_pass = 0
        self.start_time = 1558051385
        self.timestamp = 1558051385

    def cleanup(self):
        pass

    @classmethod
    def load(cls, uid):
        return cls.start(uid=uid) if uid > 0 else None

    def report(self, force=False):
        pass

    @classmethod
    def start(cls, uid=None):
        status = FakeReduceStatus()
        status.uid = 123 if uid is None else uid
        return status


class FakeTarget(object):
    "Stub to fake parts of grizzly.target.Target needed for testing the reduce loop"
    RESULT_NONE = 0
    RESULT_FAILURE = 1
    RESULT_IGNORED = 2

    def __init__(self, *args, **kwds):
        self.rl_reset = 10
        self.closed = True
        self.binary = ""
        self.forced_close = os.getenv("GRZ_FORCED_CLOSE", "1").lower() not in ("false", "0")
        self.prefs = None
        self.rl_countdown = 0
        self.use_valgrind = False

        self._calls = {
            "check_relaunch": 0,
            "cleanup": 0,
            "close": 0,
            "detect_failure": 0,
            "is_idle": 0,
            "launch": 0,
            "save_logs": 0,
        }
        self._is_healthy = False

        class FakeMonitor(object):  # pylint: disable=too-few-public-methods

            @staticmethod
            def is_healthy():
                return self._is_healthy

        self.monitor = FakeMonitor()

    def save_logs(self, dest, **kwds):
        self._calls["save_logs"] += 1
        with open(os.path.join(dest, "log_stdout.txt"), "w") as log_fp:
            log_fp.write("STDOUT")
        with open(os.path.join(dest, "log_stderr.txt"), "w") as log_fp:
            log_fp.write("Assertion failure: bad thing happened, at test.c:123")

    def is_idle(self, *args, **kwds):
        self._calls["is_idle"] += 1
        return False

    def launch(self, *args, **kwds):
        self._calls["launch"] += 1

    def check_relaunch(self):
        self._calls["check_relaunch"] += 1

    def close(self):
        self._calls["close"] += 1

    def cleanup(self):
        self._calls["cleanup"] += 1

    def detect_failure(self, *args, **kwds):
        self._calls["detect_failure"] += 1
        return Target.RESULT_FAILURE

    @property
    def expect_close(self):
        return self.rl_countdown < 1 and not self.forced_close

    def step(self):
        self.rl_countdown -= 1


def create_target_binary(target, tmp_path):
    (tmp_path / "firefox.fuzzmanagerconf").write_text(
        "[Main]\n"
        "platform = x86-64\n"
        "product = mozilla-central\n"
        "os = linux\n"
    )
    target.binary = str(tmp_path / "firefox")


class TestReductionJob(reduce.ReductionJob):
    """Stub to fake parts of grizzly.reduce.ReductionJob needed for testing the reduce loop"""
    __slots__ = []

    def __init__(self, tmp_path, create_binary=True, testcase_cache=False, skip_analysis=True):
        super(TestReductionJob, self).__init__([], FakeTarget(), 60, False, False, 0, 1, 1, 3, 25,
                                               FakeReduceStatus(),
                                               testcase_cache=testcase_cache,
                                               skip_analysis=skip_analysis)
        if create_binary:
            create_target_binary(self.target, tmp_path)

    def lithium_init(self):
        pass

    def close(self, *_args, **_kwds):
        super(TestReductionJob, self).close(keep_temp=False)

    def _run(self, testcase, temp_prefix):
        result_logs = temp_prefix + "_logs"
        os.mkdir(result_logs)
        self.target.save_logs(result_logs)
        testcase.duration = 0.1
        with open(self.reduce_file) as fp:
            return "required" in fp.read()

    def lithium_cleanup(self):
        pass


class TestMainReductionJob(TestReductionJob):
    __slots__ = []

    def __init__(self, *_args, **_kwds):
        super(TestMainReductionJob, self).__init__(None, create_binary=False)


# this is the same as TestReductionJob, but for CrashReductionJob
class TestMainCrashReductionJob(TestMainReductionJob, crash.CrashReductionJob):
    """Stub to fake parts of grizzly.crash.CrashReductionJob needed for testing the reduce loop"""
    __slots__ = []


class TestReductionJobAlt(TestReductionJob):
    """Version of TestReductionJob that only reports alternate crashes"""
    __slots__ = ['__first_run']

    def lithium_init(self):
        self.__first_run = True

    def _run(self, testcase, temp_prefix):
        result_logs = temp_prefix + "_logs"
        os.mkdir(result_logs)
        self.target.save_logs(result_logs)
        testcase.duration = 0.1
        with open(self.reduce_file) as fp:
            if "required" in fp.read():
                self.on_other_crash_found(testcase, temp_prefix)
        if self.__first_run:
            self.__first_run = False
            return True
        return False


class TestReductionJobKeepHarness(TestReductionJob):
    """Version of TestReductionJob that keeps the entire harness"""
    __slots__ = ['__init_data']

    def lithium_init(self):
        self.__init_data = None
        if os.path.basename(self.reduce_file).startswith("harness_"):
            with open(self.reduce_file) as harness_fp:
                self.__init_data = harness_fp.read()

    def _run(self, testcase, temp_prefix):
        result_logs = temp_prefix + "_logs"
        os.mkdir(result_logs)
        self.target.save_logs(result_logs)
        testcase.duration = 0.1
        if self.__init_data is not None:
            with open(self.reduce_file) as fp:
                return self.__init_data == fp.read()
        else:
            with open(self.reduce_file) as fp:
                return "required" in fp.read()


class TestReductionJobSemiReliable(TestReductionJob):
    """Version of TestReductionJob that returns interesting N times only"""
    __slots__ = ['__interesting_times', '__interesting_count', '__require_no_harness']

    def __init__(self, *args, **kwds):
        super(TestReductionJobSemiReliable, self).__init__(*args, **kwds)
        self.__interesting_times = 0
        self.__interesting_count = 0
        self.__require_no_harness = False

    def test_set_n(self, n, require_no_harness=False):
        self.__interesting_times = n
        self.__interesting_count = 0
        self.__require_no_harness = require_no_harness

    def _run(self, testcase, temp_prefix):
        result_logs = temp_prefix + "_logs"
        os.mkdir(result_logs)
        self.target.save_logs(result_logs)
        testcase.duration = 0.1
        if self.__require_no_harness and not self._no_harness:
            return False
        self.__interesting_count += 1
        return self.__interesting_count <= self.__interesting_times



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import re

__all__ = ("ReductionJob",)


def testcase_contents(path="."):
    for dir_name, _, dir_files in os.walk(path):
        arc_path = os.path.relpath(dir_name, path)
        # skip tmp folders
        if re.match(r"^tmp.+$", arc_path.split(os.sep, 1)[0]) is not None:
            continue
        for file_name in dir_files:
            # skip core files
            if re.match(r"^core.\d+$", file_name) is not None:
                continue
            if arc_path == ".":
                yield file_name
            else:
                yield os.path.join(arc_path, file_name)


from .reduce import ReductionJob  # noqa pylint: disable=wrong-import-position



# coding=utf-8
"""
SapphireJob unit tests
"""
# pylint: disable=protected-access

import platform

import pytest

from .sapphire_job import SapphireJob
from .server_map import Resource, ServerMap
from .status_codes import SERVED_ALL, SERVED_NONE, SERVED_REQUEST


def test_sapphire_job_01(tmp_path):
    """test creating an empty SapphireJob"""
    job = SapphireJob(str(tmp_path))
    assert not job.forever
    assert job.status == SERVED_ALL
    assert job.check_request("") is None
    assert job.check_request("test") is None
    assert job.check_request("test/test/") is None
    assert job.check_request("test/../../") is None
    assert not job.is_forbidden(str(tmp_path))
    assert not job.is_forbidden(str(tmp_path / "missing_file"))
    assert job.pending == 0
    assert not job.is_complete()
    assert job.remove_pending("no_file.test")
    job.finish()
    assert not any(job.served)
    assert job.is_complete()

def test_sapphire_job_02(tmp_path):
    """test SapphireJob two required files and one optional file"""
    opt_path = tmp_path / "opt_file.txt"
    opt_path.write_bytes(b"a")
    req1_path = tmp_path / "req_file_1.txt"
    req1_path.write_bytes(b"a")
    (tmp_path / "test").mkdir()
    req2_path = tmp_path / "test" / "req_file_2.txt"
    req2_path.write_bytes(b"a")
    job = SapphireJob(str(tmp_path), optional_files=[opt_path.name])
    assert job.status == SERVED_NONE
    assert not job.is_complete()
    resource = job.check_request("req_file_1.txt")
    assert resource.required
    assert resource.target == str(tmp_path / "req_file_1.txt")
    assert resource.type == Resource.URL_FILE
    assert not job.is_forbidden(str(req1_path))
    assert not job.remove_pending("no_file.test")
    assert job.pending == 2
    assert not job.remove_pending(str(req1_path))
    assert job.status == SERVED_REQUEST
    assert job.pending == 1
    assert job.remove_pending(str(req2_path))
    assert job.status == SERVED_ALL
    assert job.pending == 0
    assert job.remove_pending(str(req1_path))
    resource = job.check_request("opt_file.txt")
    assert not resource.required
    assert resource.target == str(tmp_path / "opt_file.txt")
    assert resource.type == Resource.URL_FILE
    assert job.remove_pending(str(opt_path))
    job.finish()
    assert job.is_complete()

def test_sapphire_job_03(tmp_path):
    """test SapphireJob redirects"""
    smap = ServerMap()
    smap.set_redirect("one", "somefile.txt", required=False)
    smap.set_redirect("two", "reqfile.txt")
    job = SapphireJob(str(tmp_path), server_map=smap)
    assert job.status == SERVED_NONE
    resource = job.check_request("one")
    assert resource.type == Resource.URL_REDIRECT
    resource = job.check_request("two?q=123")
    assert resource is not None
    assert resource.type == Resource.URL_REDIRECT
    assert job.pending == 1
    assert job.remove_pending("two")
    assert job.pending == 0

def test_sapphire_job_04(mocker, tmp_path):
    """test SapphireJob includes"""
    srv_root = tmp_path / "root"
    srv_include = tmp_path / "test"
    srv_include_2 = tmp_path / "test_2"
    srv_include_nested = srv_include / "nested"
    srv_root.mkdir()
    srv_include.mkdir()
    srv_include_2.mkdir()
    srv_include_nested.mkdir()
    test_1 = srv_root / "req_file.txt"
    test_1.write_bytes(b"a")
    inc_1 = srv_include / "test_file.txt"
    inc_1.write_bytes(b"b")
    nst_1 = srv_include_nested / "nested_file.txt"
    nst_1.write_bytes(b"c")
    inc_2 = srv_include_2 / "test_file_2.txt"
    inc_2.write_bytes(b"d")
    # stub out ServerMap._check_url() because it is too
    # restrictive to allow testing of some functionality
    mocker.patch.object(ServerMap, "_check_url", side_effect=lambda x: x)
    smap = ServerMap()
    smap.set_include("testinc", str(srv_include))
    # add manually to avoid sanity checks in ServerMap.set_include()
    smap.include["testinc/fakedir"] = Resource(Resource.URL_INCLUDE, str(srv_include))
    smap.include["testinc/1/2/3"] = Resource(Resource.URL_INCLUDE, str(srv_include))
    smap.include[""] = Resource(Resource.URL_INCLUDE, str(srv_include))
    smap.set_include("testinc/inc2", str(srv_include_2))
    job = SapphireJob(str(srv_root), server_map=smap)
    assert job.status == SERVED_NONE
    # test includes that map to 'srv_include'
    for incl, inc_path in smap.include.items():
        if inc_path != str(srv_include):  # only check 'srv_include' mappings
            continue
        resource = job.check_request("/".join([incl, "test_file.txt"]))
        assert resource.type == Resource.URL_INCLUDE
        assert resource.target == str(inc_1)
    # test nested include path pointing to a different include
    resource = job.check_request("testinc/inc2/test_file_2.txt?q=123")
    assert resource.type == Resource.URL_INCLUDE
    assert resource.target == str(inc_2)
    # test redirect root without leading '/'
    resource = job.check_request("test_file.txt")
    assert resource.type == Resource.URL_INCLUDE
    assert resource.target == str(srv_include / "test_file.txt")
    # test redirect with file in a nested directory
    resource = job.check_request("/".join(["testinc", "nested", "nested_file.txt"]))
    assert resource.type == Resource.URL_INCLUDE
    assert resource.target == str(nst_1)
    assert not job.is_forbidden(str(srv_root / ".." / "test" / "test_file.txt"))
    assert not job.is_forbidden(str(srv_include / ".." / "root" / "req_file.txt"))

def test_sapphire_job_05(tmp_path):
    """test SapphireJob.check_request() with tricky includes"""
    srv_root = tmp_path / "root"
    srv_root.mkdir()
    req = srv_root / "req_file.txt"
    req.write_bytes(b"a")
    inc_dir = tmp_path / "inc"
    inc_dir.mkdir()
    (inc_dir / "sub").mkdir()
    inc_file1 = inc_dir / "sub" / "include.js"
    inc_file1.write_bytes(b"a")
    inc_file2 = inc_dir / "test_inc.html"
    inc_file2.write_bytes(b"a")
    # test url matching part of the file name
    smap = ServerMap()
    smap.include["inc"] = Resource(Resource.URL_INCLUDE, str(inc_dir))
    job = SapphireJob(str(srv_root), server_map=smap)
    resource = job.check_request("inc/sub/include.js")
    assert resource.type == Resource.URL_INCLUDE
    assert resource.target == str(inc_file1)
    # test checking only the include url
    assert job.check_request("inc") is None
    # file and include file collision (files should always win)
    smap.include.clear()
    inc_a = inc_dir / "a.bin"
    inc_a.write_bytes(b"a")
    file_a = srv_root / "a.bin"
    file_a.write_bytes(b"a")
    smap.include["/"] = Resource(Resource.URL_INCLUDE, str(inc_dir))
    resource = job.check_request("a.bin")
    assert resource.type == Resource.URL_FILE
    assert resource.target == str(file_a)
    # inc and inc subdir collision
    # TODO: This can fail. How do we detect or support it?
    #smap.include.clear()
    #(inc_dir / "c").mkdir()
    #inc_c_d = (inc_dir / "c" / "d.bin")
    #inc_c_d.write_bytes(b"a")
    #inc_d = (inc_dir / "d.bin")
    #inc_d.write_bytes(b"a")
    #smap.include["c"] = Resource(Resource.URL_INCLUDE, str(inc_dir))
    #smap.include[""] = Resource(Resource.URL_INCLUDE, str(inc_dir / "c"))
    #resource = job.check_request("c/d.bin")
    #assert resource.type == Resource.URL_INCLUDE
    #assert resource.target == str(inc_c_d)

def test_sapphire_job_06(tmp_path):
    """test SapphireJob dynamic"""
    smap = ServerMap()
    smap.set_dynamic_response("cb1", lambda: 0, mime_type="mime_type")
    smap.set_dynamic_response("cb2", lambda: 1)
    job = SapphireJob(str(tmp_path), server_map=smap)
    assert job.status == SERVED_ALL
    assert job.pending == 0
    resource = job.check_request("cb1")
    assert resource.type == Resource.URL_DYNAMIC
    assert callable(resource.target)
    assert isinstance(resource.mime, str)
    resource = job.check_request("cb2?q=123")
    assert resource is not None
    assert resource.type == Resource.URL_DYNAMIC
    assert callable(resource.target)
    assert isinstance(resource.mime, str)

def test_sapphire_job_07(tmp_path):
    """test accessing forbidden files"""
    srv_root = tmp_path / "root"
    srv_root.mkdir()
    test_1 = srv_root / "req_file.txt"
    test_1.write_bytes(b"a")
    no_access = tmp_path / "no_access.txt"
    no_access.write_bytes(b"a")
    job = SapphireJob(str(srv_root))
    assert job.status == SERVED_NONE
    assert job.pending == 1
    resource = job.check_request("../no_access.txt")
    assert resource.target == str(no_access)
    assert resource.type == Resource.URL_FILE
    assert not job.is_forbidden(str(test_1))
    assert job.is_forbidden(str(srv_root / "../no_access.txt"))

@pytest.mark.skipif(platform.system() == "Windows",
                    reason="Unsupported on Windows")
def test_sapphire_job_08(tmp_path):
    """test SapphireJob with file names containing invalid characters"""
    test_file = tmp_path / "test.txt"
    test_file.write_bytes(b"a")
    (tmp_path / "?_2.txt").write_bytes(b"a")
    job = SapphireJob(str(tmp_path))
    assert job.status == SERVED_NONE
    assert job.pending == 1
    assert job.check_request("test.txt").target == str(test_file)

def test_sapphire_job_09():
    """test SapphireJob with missing directory"""
    with pytest.raises(OSError):
        SapphireJob("missing")

def test_sapphire_job_10(tmp_path):
    """test SapphireJob.increment_served() and SapphireJob.served"""
    job = SapphireJob(str(tmp_path))
    assert not any(job.served)
    job.increment_served(str(tmp_path / "file.bin"))
    assert "file.bin" in job.served
    job.increment_served("/some/include/path/inc.bin")
    assert "/some/include/path/inc.bin" in job.served



# coding: utf-8
"""
Loki fuzzing library
"""
from __future__ import print_function
import argparse
import logging
import logging.handlers
import os
import random
import struct
import shutil
import tempfile
import time


__author__ = "Tyson Smith"
LOG = logging.getLogger("loki")


class Loki(object):

    def __init__(self, aggression=0.0):
        self.aggr = min(max(aggression, 0.0), 1.0)

    @staticmethod
    def _fuzz_data(in_data, byte_order=None):
        data_size = len(in_data)
        if data_size == 1:
            pack_unit = "B"
            mask = 0xFF
        elif data_size == 2:
            pack_unit = "H"
            mask = 0xFFFF
        elif data_size == 4:
            pack_unit = "I"
            mask = 0xFFFFFFFF
        else:
            raise RuntimeError("Unsupported data size: %d" % data_size)

        if byte_order is None:
            byte_order = "<" if random.getrandbits(5) else ">"  # prefer little-endian
        elif byte_order not in (">", "<"):
            raise RuntimeError("Unsupported byte order %r" % byte_order)
        pack_unit = "".join([byte_order, pack_unit])

        fuzz_op = random.randint(0, 4)
        if fuzz_op == 0:  # boundary
            out_data = (2**random.randint(2, (data_size * 8))) + random.randint(-2, 2)
        elif fuzz_op == 1:  # arithmetic
            out_data = struct.unpack(pack_unit, in_data)[0] + random.randint(-10, 10)
        elif fuzz_op == 2:  # interesting byte, short or int
            out_data = random.choice((0, 1, int(mask / 2), int(mask / 2) + 1, mask))
        elif fuzz_op == 3:  # random byte, short or int
            out_data = random.getrandbits(32)
        elif fuzz_op == 4 and data_size == 1:  # toggle
            out_data = struct.unpack(pack_unit, in_data)[0] ^ (2 ** random.getrandbits(3))
        elif fuzz_op == 4 and data_size == 2:  # multiple of a data size
            if random.getrandbits(1):
                out_data = random.randint(1, 0x1FFF) * 8
            else:
                out_data = random.randint(1, 0xFFF) * 16
        elif fuzz_op == 4 and data_size == 4:  # multiple of a data size
            fuzz_op = random.randint(0, 2)
            if fuzz_op == 0:
                out_data = random.randint(1, 0x1FFFFFFF) * 8
            elif fuzz_op == 1:
                out_data = random.randint(1, 0xFFFFFFF) * 16
            elif fuzz_op == 2:
                out_data = random.randint(1, 0x7FFFFFF) * 32

        return struct.pack(pack_unit, out_data & mask)

    def _fuzz(self, tgt_fp):
        tgt_fp.seek(0, os.SEEK_END)
        length = tgt_fp.tell()
        if length < 1:
            raise RuntimeError("Zero length file cannot be fuzzed.")

        # minimum number of max passes should be 1
        max_passes = max(int(round(length * self.aggr)), 1)
        fuzz_passes = random.randint(1, max_passes)
        LOG.debug("%d of a possible %d fuzz passes will be performed", fuzz_passes, max_passes)

        max_bytes = min(length, 2) if length < 4 else 4
        for _ in range(fuzz_passes):
            if max_bytes > 1 and not random.getrandbits(4):  # 6.25%
                fuzz_size = max_bytes >> 1 if random.getrandbits(1) else max_bytes
            else:
                fuzz_size = 1
            target = random.randint(0, length-fuzz_size)

            tgt_fp.seek(target)
            out_data = self._fuzz_data(tgt_fp.read(fuzz_size))
            tgt_fp.seek(target)
            tgt_fp.write(out_data)

    def fuzz_data(self, data):
        assert isinstance(data, bytes)
        # open a temp file in memory for fuzzing
        with tempfile.SpooledTemporaryFile(max_size=0x800000, mode="r+b") as tmp_fp:
            tmp_fp.write(data)
            self._fuzz(tmp_fp)
            tmp_fp.seek(0)
            return tmp_fp.read()

    def fuzz_file(self, in_file, count, ext=None, out_dir=None):
        start_time = time.time()
        LOG.info("Starting Loki @ %s", time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))

        # Analyze input test case
        LOG.info("Target template is %r", os.path.abspath(in_file))
        try:
            length = os.path.getsize(in_file)
        except OSError:
            LOG.error("%r does not exists!", in_file)
            return False

        if length < 1:
            LOG.error("Input must be at least 1 byte long")
            return False

        LOG.info("Template size in bytes is %d", length)
        if ext is None:
            ext = os.path.splitext(in_file)[1]

        # Checking output directory
        if out_dir is None:
            out_dir = tempfile.mkdtemp(
                prefix=time.strftime("loki_%m%d%H%M_", time.localtime()),
                dir=".")
        elif not os.path.isdir(out_dir):
            os.mkdir(out_dir)
        LOG.info("Output directory is %s", os.path.abspath(out_dir))

        LOG.info("Generating %s fuzzed test cases...", count)
        for i in range(count):
            out_file = os.path.join(out_dir, "".join(("%06d_fuzzed" % i, ext)))
            shutil.copy(in_file, out_file)

            with open(out_file, "r+b") as out_fp:
                self._fuzz(out_fp)

        finish_time = time.time()-start_time
        LOG.info("Total run time %gs", finish_time)
        if count > 0:
            LOG.info("About %gs per file", finish_time/count)

        return True

    @staticmethod
    def splice_data(data_chunks):
        if len(data_chunks) not in (1, 2):
            return None  # one or two data blobs are required (one truncates)

        blob_pass = 1
        with tempfile.SpooledTemporaryFile(max_size=0x800000, mode="r+b") as tmp_fp:
            for chunk in data_chunks:
                length = len(chunk)

                if length < 1:
                    return None  # not enough data chunks to work with

                target = random.randint(0, length-1)

                if blob_pass == 1:
                    tmp_fp.write(chunk[:target])
                elif blob_pass == 2:
                    tmp_fp.write(chunk[target:])

                blob_pass += 1

            tmp_fp.seek(0)
            return tmp_fp.read()


def main():
    parser = argparse.ArgumentParser(description="Loki fuzzing library")
    parser.add_argument(
        "input",
        help="Output will be generated based on this file")
    parser.add_argument(
        "-a", "--aggression", default=0.001, type=float,
        help="Controls how much fuzzing is done on the output file")
    parser.add_argument(
        "-c", "--count", default=100, type=int,
        help="Number test cases to generate")
    parser.add_argument(
        "-q", "--quiet", default=False, action="store_true",
        help="Display limited output")
    parser.add_argument(
        "-o", "--output", default=None,
        help="Output directory for fuzzed test cases")

    args = parser.parse_args()

    # setup logging
    hnd = logging.StreamHandler()
    hnd.setFormatter(logging.Formatter("[%(levelname).1s] %(message)s"))
    LOG.addHandler(hnd)
    LOG.setLevel(logging.INFO if not args.quiet else logging.ERROR)

    loki = Loki(aggression=args.aggression)
    try:
        loki.fuzz_file(args.input, args.count, out_dir=args.output)
    except KeyboardInterrupt:
        print("Ctrl+C detected.")



import importlib
import logging
import os
import sys
import traceback

from grizzly.common import Adapter

log = logging.getLogger("grizzly")  # pylint: disable=invalid-name

__all__ = ("get", "load", "names")
__adapters__ = dict()

def get(name):
    return __adapters__.get(name.lower(), None)

def load(path=None, skip_failures=True):
    assert not __adapters__, "adapters have already been loaded"
    if path is None:
        path = os.path.dirname(__file__)
    path = os.path.abspath(path)
    log.debug("loading adapters from %r", path)
    sys.path.append(path)
    for sub in os.listdir(path):
        if not os.path.isfile(os.path.join(path, sub, "__init__.py")):
            continue
        log.debug("scanning %r", sub)
        try:
            lib = importlib.import_module(sub)
        except Exception:  # pylint: disable=broad-except
            if not skip_failures:
                raise
            exc_type, exc_obj, exc_tb = sys.exc_info()
            tbinfo = traceback.extract_tb(exc_tb)[-1]
            log.debug("raised %s: %s (%s:%d)", exc_type.__name__, exc_obj, tbinfo[0], tbinfo[1])
            continue
        for clsname in dir(lib):
            cls = getattr(lib, clsname)
            if isinstance(cls, type) and issubclass(cls, Adapter):
                if clsname == "Adapter":
                    continue
                log.debug("sanity checking %r", clsname)
                if not isinstance(cls.NAME, str):
                    raise RuntimeError(
                        "%s.NAME must be 'str' not %r" % (cls.__name__, type(cls.NAME).__name__))
                if cls.NAME.lower() != cls.NAME:
                    raise RuntimeError(
                        "%s.NAME %r must be lowercase" % (cls.__name__, cls.NAME))
                if cls.NAME in __adapters__:
                    raise RuntimeError(
                        "Name collision! %r is used by %r and %r" % (
                            cls.NAME,
                            __adapters__[cls.NAME].__name__,
                            cls.__name__))
                __adapters__[cls.NAME] = cls
        else:
            log.debug("ignored %r", sub)
    log.debug("%d adapters loaded", len(__adapters__))

def names():
    return __adapters__.keys()



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from logging import getLogger
from os.path import isdir
from shutil import rmtree
from tempfile import mkdtemp
from time import sleep, time

from .common import grz_tmp, Report, Runner, Status, TestFile
from .target import TargetLaunchError


__all__ = ("SessionError", "LogOutputLimiter", "Session")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


log = getLogger("grizzly")  # pylint: disable=invalid-name


class SessionError(Exception):
    """The base class for exceptions raised by Session"""


class LogOutputLimiter(object):
    __slots__ = ("_delay", "_iterations", "_launches", "_multiplier", "_time", "_verbose")

    def __init__(self, delay=300, delta_multiplier=2, verbose=False):
        self._delay = delay  # maximum time delay between output
        self._iterations = 1  # next iteration to trigger output
        self._launches = 1  # next launch to trigger output
        self._multiplier = delta_multiplier  # rate to decrease output (iterations)
        self._time = time()
        self._verbose = verbose  # always output

    def ready(self, cur_iter, launches):
        # calculate if a status line should be output
        if self._verbose:
            return True
        ready = False
        if cur_iter >= self._iterations:
            ready = True
            self._iterations *= self._multiplier
        elif launches >= self._launches:
            ready = True
        elif time() - self._delay >= self._time:
            ready = True
        if ready:
            self._time = time()
            self._launches = launches + 1
        return ready


class Session(object):
    DISPLAY_VERBOSE = 0  # display status every iteration
    DISPLAY_NORMAL = 1  # quickly reduce the amount of output
    EXIT_SUCCESS = 0
    EXIT_ERROR = 1
    EXIT_ABORT = 3
    EXIT_LAUNCH_FAILURE = 7
    TARGET_LOG_SIZE_WARN = 0x1900000  # display warning when target log files exceed limit (25MB)

    __slots__ = ("adapter", "coverage", "iomanager", "reporter", "server", "status", "target")

    def __init__(self, adapter, iomanager, reporter, server, target, coverage=False):
        self.adapter = adapter
        self.coverage = coverage
        self.iomanager = iomanager
        self.reporter = reporter
        self.server = server
        self.status = Status.start()
        self.target = target

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        self.status.cleanup()

    def display_status(self, log_limiter):
        if self.adapter.remaining is not None:
            log.info(
                "[I%04d-L%02d-R%02d] %s",
                self.status.iteration,
                self.adapter.remaining,
                self.status.results,
                self.status.test_name)
        elif log_limiter.ready(self.status.iteration, self.target.monitor.launches):
            log.info("I%04d-R%02d ", self.status.iteration, self.status.results)

    def generate_testcase(self):
        log.debug("calling iomanager.create_testcase()")
        test = self.iomanager.create_testcase(self.adapter.NAME)
        log.debug("calling self.adapter.generate()")
        self.adapter.generate(test, self.iomanager.server_map)
        self.status.test_name = test.input_fname
        if self.target.prefs is not None:
            # TODO: this can likely be improved
            test.add_meta(TestFile.from_file(self.target.prefs, "prefs.js"))
        return test

    def report_result(self):
        # create working directory for target logs
        result_logs = mkdtemp(prefix="logs_", dir=grz_tmp("logs"))
        self.target.save_logs(result_logs)
        report = Report.from_path(result_logs)
        crash_info = report.crash_info(self.target.binary)
        short_sig = crash_info.createShortSignature()
        log.info("Result: %s (%s:%s)", short_sig, report.major[:8], report.minor[:8])
        # order test cases newest to oldest
        self.iomanager.tests.reverse()
        self.reporter.submit(self.iomanager.tests, report=report)
        if isdir(result_logs):
            rmtree(result_logs)

    def run(self, ignore, iteration_limit=None, display_mode=DISPLAY_NORMAL):
        log_limiter = LogOutputLimiter(verbose=display_mode == self.DISPLAY_VERBOSE)
        runner = Runner(self.server, self.target)

        def _dyn_close():  # pragma: no cover
            if self.target.monitor.is_healthy():
                # delay to help catch window close/shutdown related crashes
                sleep(0.1)
                self.target.close()
            return b"<h1>Close Browser</h1>"
        self.iomanager.server_map.set_dynamic_response(
            "grz_close_browser",
            _dyn_close,
            mime_type="text/html")

        while True:
            self.status.report()
            self.status.iteration += 1

            if self.target.closed:
                # (re-)launch target
                self.iomanager.purge_tests()
                self.adapter.pre_launch()
                if self.iomanager.harness is None:
                    # harness is not in use, open the test case
                    location = runner.location(
                        "/grz_current_test",
                        self.server.port)
                else:
                    # harness is in use, open it and it will open the test case
                    location = runner.location(
                        "/grz_harness",
                        self.server.port,
                        close_after=self.target.rl_reset,
                        forced_close=self.target.forced_close,
                        timeout=self.adapter.TEST_DURATION)
                log.info("Launching target")
                try:
                    runner.launch(location, max_retries=3, retry_delay=0)
                except TargetLaunchError:
                    # this result likely has nothing to do with Grizzly
                    self.status.results += 1
                    log.error("Target launch error. Check browser logs for details.")
                    self.report_result()
                    raise
            self.target.step()

            # create and populate a test case
            current_test = self.generate_testcase()
            # display status
            self.display_status(log_limiter=log_limiter)

            # run test case
            runner.run(ignore, self.iomanager.server_map, current_test, coverage=self.coverage)
            # adapter callbacks
            if runner.timeout:
                log.debug("calling self.adapter.on_timeout()")
                self.adapter.on_timeout(current_test, runner.served)
            else:
                log.debug("calling self.adapter.on_served()")
                self.adapter.on_served(current_test, runner.served)
            # update test case
            if runner.result != runner.ERROR:
                if not runner.served:
                    # this can happen if the target crashes between serving test cases
                    log.info("Ignoring test case since nothing was served")
                    self.iomanager.tests.pop().cleanup()
                elif self.adapter.IGNORE_UNSERVED:
                    log.debug("removing unserved files from the test case")
                    current_test.purge_optional(runner.served)
            # process results
            if runner.result == runner.FAILED:
                self.status.results += 1
                log.debug("result detected")
                self.report_result()
            elif runner.result == runner.IGNORED:
                self.status.ignored += 1
                log.info("Ignored (%d)", self.status.ignored)
            elif runner.result == runner.ERROR:
                log.error("Test case was not served")
                if not current_test.contains(current_test.landing_page):
                    log.warning("Test case is missing landing page")
                if self.status.iteration < 3:
                    # since this is the first few iteration something is
                    # likely wrong with the Target (caching?) or the Adapter
                    raise SessionError("Please check Adapter and Target")

            # trigger relaunch by closing the browser if needed
            self.target.check_relaunch()

            if self.adapter.remaining is not None and self.adapter.remaining < 1:
                # all test cases have been replayed
                log.info("Replay Complete")
                break

            if iteration_limit is not None and self.status.iteration == iteration_limit:
                log.info("Hit iteration limit")
                break

            # warn about large browser logs
            self.status.log_size = self.target.log_size()
            if self.status.log_size > self.TARGET_LOG_SIZE_WARN:
                log.warning("Large browser logs: %dMBs", (self.status.log_size / 0x100000))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from argparse import ArgumentParser
from logging import basicConfig, DEBUG, INFO
from os.path import isdir

from .core import Sapphire


def configure_logging(log_level):
    if log_level == DEBUG:
        log_fmt = "%(levelname).1s %(name)s [%(asctime)s] %(message)s"
    else:
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt="%Y-%m-%d %H:%M:%S", level=log_level)

def parse_args(argv=None):
    # log levels for console logging
    level_map = {"DEBUG": DEBUG, "INFO": INFO}
    parser = ArgumentParser()
    parser.add_argument(
        "path",
        help="Specify a directory to act as wwwroot")
    parser.add_argument(
        "--log-level", default="INFO",
        help="Configure console logging. Options: %s (default: %%(default)s)" %
        ", ".join(k for k, v in sorted(level_map.items(), key=lambda x: x[1])))
    parser.add_argument(
        "--port", type=int,
        help="Specify a port to bind to (default: random)")
    parser.add_argument(
        "--remote", action="store_true",
        help="Allow connections from addresses other than 127.0.0.1")
    parser.add_argument(
        "--timeout", type=int,
        help="Duration in seconds to serve before exiting. Default run forever.")
    args = parser.parse_args(argv)
    # sanity check
    if not isdir(args.path):
        parser.error("Path does not exist %r" % (args.path,))
    if args.timeout is not None and args.timeout <= 0:
        parser.error("Specified timeout must be greater than 0")
    log_level = level_map.get(args.log_level.upper(), None)
    if log_level is None:
        parser.error("Invalid log-level %r" % (args.log_level,))
    args.log_level = log_level
    return args

ARGS = parse_args()
configure_logging(ARGS.log_level)
Sapphire.main(ARGS)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access

import json
import re
import os

import pytest

from .storage import TestCase, TestFile, TestCaseLoadFailure, TestFileExists


def test_testcase_01(tmp_path):
    """test empty TestCase"""
    l_page = "land.html"
    r_page = "redirect.html"
    adpt_name = "test-adapter"
    with TestCase(l_page, r_page, adpt_name) as tcase:
        assert tcase.landing_page == l_page
        assert tcase.redirect_page == r_page
        assert tcase.adapter_name == adpt_name
        assert tcase.duration is None
        assert tcase.data_size == 0
        assert tcase.input_fname is None
        assert not tcase.env_vars
        assert not tcase._existing_paths
        assert not tcase._files.meta
        assert not tcase._files.optional
        assert not tcase._files.required
        assert not tcase.contains("no_file")
        assert not any(tcase.optional)
        tcase.dump(str(tmp_path))
        assert not any(tmp_path.glob("*"))
        tcase.dump(str(tmp_path), include_details=True)
        assert (tmp_path / "test_info.json").is_file()

def test_testcase_02(tmp_path):
    """test TestCase with TestFiles"""
    tcase = TestCase("land_page.html", "redirect.html", "test-adapter", input_fname="testinput.bin")
    try:
        in_file = tmp_path / "testfile1.bin"
        in_file.write_bytes(b"test_req")
        tcase.add_from_file(str(in_file))
        assert tcase.data_size == 8
        with pytest.raises(TestFileExists, match="'testfile1.bin' exists in test"):
            tcase.add_from_file(str(in_file), file_name="testfile1.bin")
        with pytest.raises(TestFileExists, match="'testfile1.bin' exists in test"):
            tcase.add_from_data("test", "testfile1.bin")
        tcase.add_from_data("test_nreq", "nested/testfile2.bin", required=False)
        tcase.add_from_data("test_blah", "/testfile3.bin")
        tcase.add_from_data("test_windows", "\\\\dir\\file.bin")
        assert tcase.contains("testfile1.bin")
        opt_files = list(tcase.optional)
        assert os.path.join("nested", "testfile2.bin") in opt_files
        assert len(opt_files) == 1
        tcase.dump(str(tmp_path), include_details=True)
        assert (tmp_path / "nested").is_dir()
        test_info = json.loads((tmp_path / "test_info.json").read_text())
        assert test_info["adapter"] == "test-adapter"
        assert test_info["input"] == "testinput.bin"
        assert test_info["target"] == "land_page.html"
        assert isinstance(test_info["env"], dict)
        assert in_file.read_bytes() == b"test_req"
        assert (tmp_path / "nested" / "testfile2.bin").read_bytes() == b"test_nreq"
        assert (tmp_path / "testfile3.bin").read_bytes() == b"test_blah"
        assert (tmp_path / "dir" / "file.bin").read_bytes() == b"test_windows"
    finally:
        tcase.cleanup()

def test_testcase_03(tmp_path):
    """test TestCase.add_meta()"""
    with TestCase("land_page.html", "redirect.html", "test-adapter") as tcase:
        dmp_path = tmp_path / "dmp_test"
        dmp_path.mkdir()
        meta_file = dmp_path / "metafile.bin"
        meta_data = b"foobar"
        tcase.add_meta(TestFile.from_data(meta_data, meta_file.name))
        tcase.dump(str(dmp_path), include_details=True)
        assert tcase.data_size == 6
        assert meta_file.is_file()
        assert meta_file.read_bytes() == meta_data

def test_testcase_04(tmp_path):
    """test TestCase.add_environ_var() and TestCase.env_vars"""
    with TestCase("land_page.html", "redirect.html", "test-adapter") as tcase:
        tcase.add_environ_var("TEST_ENV_VAR", "1")
        assert len(list(tcase.env_vars)) == 1
        tcase.add_environ_var("TEST_NONE", None)
        assert len(tcase.env_vars) == 2
        dmp_path = tmp_path / "dmp_test"
        dmp_path.mkdir()
        tcase.dump(str(dmp_path), include_details=True)
        data = json.loads((dmp_path / "test_info.json").read_text())
        assert "env" in data
        assert data["env"]["TEST_ENV_VAR"] == "1"
        assert data["env"]["TEST_NONE"] is None

def test_testcase_05(tmp_path):
    """test TestCase.purge_optional()"""
    with TestCase("land_page.html", "redirect.html", "test-adapter") as tcase:
        tcase.add_from_data("foo", "testfile1.bin")
        tcase.add_from_data("foo", "testfile2.bin", required=False)
        tcase.add_from_data("foo", "testfile3.bin", required=False)
        tcase.add_from_data("foo", "not_served.bin", required=False)
        assert len(tuple(tcase.optional)) == 3
        tcase.purge_optional(tcase.optional)
        assert len(tuple(tcase.optional)) == 3
        tcase.purge_optional(["testfile2.bin", "testfile3.bin"])
        assert len(tuple(tcase.optional)) == 2
        tcase.dump(str(tmp_path))
        assert tmp_path.glob("testfile1.bin")
        assert not any(tmp_path.glob("not_served.bin"))

def test_testcase_06():
    """test TestCase.data_size"""
    with TestCase("land_page.html", "redirect.html", "test-adapter") as tcase:
        assert tcase.data_size == 0
        tcase.add_from_data("1", "testfile1.bin", required=True)
        assert tcase.data_size == 1
        tcase.add_from_data("12", "testfile2.bin", required=False)
        assert tcase.data_size == 3
        tcase.add_meta(TestFile.from_data("123", "meta.bin"))
        assert tcase.data_size == 6

def test_testcase_07(tmp_path):
    """test TestCase.load_path() using a directory fail cases"""
    # missing test_info.json
    with pytest.raises(TestCaseLoadFailure, match="Missing 'test_info.json'"):
        TestCase.load_path(str(tmp_path))
    # invalid test_info.json
    (tmp_path / "test_info.json").write_bytes(b"X")
    with pytest.raises(TestCaseLoadFailure, match="Invalid 'test_info.json'"):
        TestCase.load_path(str(tmp_path))
    # test_info.json missing 'target' entry
    (tmp_path / "test_info.json").write_bytes(b"{}")
    with pytest.raises(TestCaseLoadFailure, match="'test_info.json' missing 'target' entry"):
        TestCase.load_path(str(tmp_path))
    # build a test case
    src_dir = (tmp_path / "src")
    src_dir.mkdir()
    (src_dir / "prefs.js").touch()
    entry_point = src_dir / "target.bin"
    entry_point.touch()
    with TestCase("target.bin", None, "test-adapter") as src:
        src.add_from_file(str(entry_point))
        src.dump(str(src_dir), include_details=True)
    # bad test_info.json 'target' entry
    entry_point.unlink()
    with pytest.raises(TestCaseLoadFailure, match="entry_point 'target.bin' not found in"):
        TestCase.load_path(str(src_dir))
    # bad test_info.json 'env' entry
    entry_point.touch()
    with TestCase("target.bin", None, "test-adapter") as src:
        src.add_environ_var("TEST_ENV_VAR", 100)
        src.dump(str(src_dir), include_details=True)
    with pytest.raises(TestCaseLoadFailure, match="'env_data' contains invalid 'env' entries"):
        TestCase.load_path(str(src_dir))

def test_testcase_08(tmp_path):
    """test TestCase.load_path() using a directory"""
    # build a valid test case
    src_dir = (tmp_path / "src")
    src_dir.mkdir()
    (src_dir / "prefs.js").touch()
    entry_point = src_dir / "target.bin"
    entry_point.touch()
    (src_dir / "optional.bin").touch()
    (src_dir / "x.bin").touch()
    nested = (tmp_path / "src" / "nested")
    nested.mkdir()
    # overlap file name in different directories
    (nested / "x.bin").touch()
    (tmp_path / "src" / "nested" / "empty").mkdir()
    with TestCase("target.bin", None, "test-adapter") as src:
        src.add_environ_var("TEST_ENV_VAR", "100")
        src.add_from_file(str(entry_point))
        src.dump(str(src_dir), include_details=True)
    # load test case from test_info.json
    with TestCase.load_path(str(src_dir)) as dst:
        assert dst.landing_page == "target.bin"
        assert "prefs.js" in (x.file_name for x in dst._files.meta)
        assert "target.bin" in (x.file_name for x in dst._files.required)
        assert "optional.bin" in (x.file_name for x in dst._files.optional)
        assert "x.bin" in (x.file_name for x in dst._files.optional)
        assert os.path.join("nested", "x.bin") in (x.file_name for x in dst._files.optional)
        assert dst.env_vars["TEST_ENV_VAR"] == "100"

def test_testcase_09(tmp_path):
    """test TestCase.load_path() using a file"""
    # invalid entry_point specified
    with pytest.raises(TestCaseLoadFailure, match="Cannot find"):
        TestCase.load_path(str(tmp_path / "missing_file"))
    # valid test case
    src_dir = (tmp_path / "src")
    src_dir.mkdir()
    (src_dir / "prefs.js").touch()
    entry_point = src_dir / "target.bin"
    entry_point.touch()
    (src_dir / "optional.bin").touch()
    # load single file test case
    with TestCase.load_path(str(entry_point)) as tcase:
        assert tcase.landing_page == "target.bin"
        assert "prefs.js" not in (x.file_name for x in tcase._files.meta)
        assert "target.bin" in (x.file_name for x in tcase._files.required)
        assert "optional.bin" not in (x.file_name for x in tcase._files.optional)
    # load full test case
    with TestCase.load_path(str(entry_point), full_scan=True, prefs=True) as tcase:
        assert tcase.landing_page == "target.bin"
        assert "prefs.js" in (x.file_name for x in tcase._files.meta)
        assert "target.bin" in (x.file_name for x in tcase._files.required)
        assert "optional.bin" in (x.file_name for x in tcase._files.optional)

def test_testcase_10(tmp_path):
    """test TestCase.load_environ()"""
    (tmp_path / "ubsan.supp").touch()
    (tmp_path / "other_file").touch()
    with TestCase("a.html", "b.html", "test-adapter") as tcase:
        tcase.load_environ(str(tmp_path), {})
        assert "UBSAN_OPTIONS" in tcase.env_vars
        assert "ubsan.supp" in tcase.env_vars["UBSAN_OPTIONS"]
        # existing *SAN_OPTIONS
        tcase.load_environ(str(tmp_path), {"UBSAN_OPTIONS": "a=1:b=2"})
        assert "UBSAN_OPTIONS" in tcase.env_vars
        assert "ubsan.supp" in tcase.env_vars["UBSAN_OPTIONS"]
        opts = re.split(r":(?![\\|/])", tcase.env_vars["UBSAN_OPTIONS"])
        assert "a=1" in opts
        assert "b=2" in opts
        assert len(opts) == 3

def test_testcase_11(tmp_path):
    """test TestCase.add_batch()"""
    include = (tmp_path / "inc_path")
    include.mkdir()
    inc_1 = (include / "file.bin")
    inc_1.write_bytes(b"a")
    (include / "nested").mkdir()
    inc_2 = (include / "nested" / "nested.js")
    inc_2.write_bytes(b"a")
    other_path = (tmp_path / "other_path")
    other_path.mkdir()
    (other_path / "no_include.bin").write_bytes(b"a")
    with TestCase("a.b", "a.b", "simple") as tcase:
        # missing directory
        tcase.add_batch("/missing/path/", tuple())
        assert not tcase._existing_paths
        # missing file
        with pytest.raises(IOError):
            tcase.add_batch(str(tmp_path), [str(tmp_path / "missing.bin")])
        assert not tcase._existing_paths
        # relative file name
        tcase.add_batch(str(include), ["file.bin"])
        assert not tcase._existing_paths
        # valid list
        tcase.add_batch(str(include), [str(inc_1), str(inc_2), str(tmp_path / "inc_path2" / "extra.bin")])
        assert tcase.contains("file.bin")
        assert tcase.contains(os.path.join("nested", "nested.js"))
        assert len(tcase._existing_paths) == 2
        # nested url
        tcase.add_batch(str(include), [str(inc_1)], prefix="test")
        assert tcase.contains(os.path.join("test", "file.bin"))
        assert len(tcase._existing_paths) == 3
        # collision
        with pytest.raises(TestFileExists, match="'file.bin' exists in test"):
            tcase.add_batch(str(include), [str(inc_1)])

def test_testfile_01():
    """test simple TestFile"""
    with pytest.raises(TypeError, match="TestFile requires a name"):
        TestFile("")
    with TestFile("test_file.txt") as tfile:
        assert tfile.file_name == "test_file.txt"
        assert not tfile._fp.closed
        assert tfile.size == 0
        tfile.close()
        assert tfile._fp.closed

def test_testfile_02(tmp_path):
    """test TestFile.write() and TestFile.dump()"""
    out_file = tmp_path / "test_file.txt"
    with TestFile(out_file.name) as tfile:
        tfile.write(b"foo")
        assert not out_file.is_file()
        tfile.dump(str(tmp_path))
        assert out_file.is_file()
        assert out_file.read_text() == "foo"
        tfile.write(b"bar")
        tfile.dump(str(tmp_path))
        assert out_file.read_text() == "foobar"

def test_testfile_03(tmp_path):
    """test TestFile.dump() file with nested path"""
    file_path = "test/dir/path/file.txt"
    with TestFile(file_path) as tfile:
        out_file = tmp_path / file_path
        tfile.write(b"foo")
        assert not out_file.is_file()
        tfile.dump(str(tmp_path))
        assert out_file.is_file()

def test_testfile_04(tmp_path):
    """test TestFile.from_data()"""
    # TODO: different encodings
    with TestFile.from_data("foo", "test_file.txt") as tfile:
        out_file = tmp_path / "test_file.txt"
        tfile.dump(str(tmp_path))
        assert out_file.is_file()
        assert out_file.read_text() == "foo"

def test_testfile_05(tmp_path):
    """test TestFile.from_file()"""
    in_file = tmp_path / "infile.txt"
    in_file.write_bytes(b"foobar")
    # check re-using filename
    with TestFile.from_file(str(in_file)) as tfile:
        assert tfile.file_name == "infile.txt"
    # check data
    with TestFile.from_file(str(in_file), file_name="outfile.txt") as tfile:
        assert tfile.file_name == "outfile.txt"
        tfile.dump(str(tmp_path))
        out_file = tmp_path / "outfile.txt"
        assert out_file.is_file()
        assert out_file.read_text() == "foobar"

def test_testfile_06(tmp_path):
    """test TestFile.clone()"""
    out_file = tmp_path / "test_file.txt"
    with TestFile(out_file.name) as tf1:
        tf1.write(b"foobar")
        try:
            tf2 = tf1.clone()
            tf2.write(b"test")
            assert tf1.file_name == tf2.file_name
            assert tf1._fp != tf2._fp
            tf2.dump(str(tmp_path))
            assert out_file.is_file()
            assert out_file.read_text() == "foobartest"
        finally:
            tf2.close()
        tf1.dump(str(tmp_path))
        assert out_file.is_file()
        assert out_file.read_text() == "foobar"

def test_testfile_07(tmp_path):
    """test TestFile.data()"""
    in_file = tmp_path / "infile.txt"
    in_file.write_bytes(b"foobar")
    with TestFile.from_file(str(in_file), file_name="outfile.txt") as tfile:
        assert tfile.data == b"foobar"



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger
import sys
import threading
import time
import traceback

from .sapphire_worker import SapphireWorker

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger("sphr_loadmgr")


class SapphireLoadManager(object):
    SHUTDOWN_DELAY = 0.5  # allow extra time before closing socket if needed

    __slots__ = ("_job", "_listener", "_socket", "_workers")

    def __init__(self, job, sock, max_workers=1):
        assert max_workers > 0
        self._job = job
        self._listener = None
        self._socket = sock
        self._workers = max_workers

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        self._job.finish()
        if self._listener is not None:
            self._listener.join()
            self._listener = None
        if not self._job.exceptions.empty():
            exc_type, exc_obj, exc_tb = self._job.exceptions.get()
            LOG.error(
                "Unexpected exception:\n%s",
                "".join(traceback.format_exception(exc_type, exc_obj, exc_tb)))
            # re-raise exception from worker once all workers are closed
            raise exc_obj

    def start(self):
        assert self._job.pending
        # create the listener thread to handle incoming requests
        listener = threading.Thread(
            target=self.listener,
            args=(self._socket, self._job, self._workers),
            kwargs={"shutdown_delay": self.SHUTDOWN_DELAY})
        # launch listener thread and handle thread errors
        for retry in reversed(range(10)):
            try:
                listener.start()
            except threading.ThreadError:
                # thread errors can be due to low system resources while fuzzing
                LOG.warning("ThreadError (listener), threads: %d", threading.active_count())
                if retry < 1:
                    raise
                time.sleep(1)
                continue
            self._listener = listener
            break

    def wait(self, timeout, continue_cb=None, poll=0.5):
        assert self._listener is not None
        if timeout > 0:
            deadline = time.time() + timeout
        else:
            deadline = None
        if continue_cb is not None and not callable(continue_cb):
            raise TypeError("continue_cb must be callable")
        # it is important to keep this loop fast because it can limit
        # the total iteration rate of Grizzly
        while not self._job.is_complete(wait=poll):
            # check for a timeout
            if deadline and deadline <= time.time():
                return False
            # check if callback returns False
            if continue_cb is not None and not continue_cb():
                LOG.debug("continue_cb() returned False")
                break
        return True

    @staticmethod
    def listener(serv_sock, serv_job, max_workers, shutdown_delay=0):
        assert max_workers > 0
        assert shutdown_delay >= 0
        worker_pool = list()
        pool_size = 0
        LOG.debug("starting listener")
        try:
            while not serv_job.is_complete():
                if not serv_job.accepting.wait(0.05):
                    continue
                worker = SapphireWorker.launch(serv_sock, serv_job)
                if worker is not None:
                    worker_pool.append(worker)
                    pool_size += 1
                # manage worker pool
                if pool_size >= max_workers:
                    LOG.debug("pool size: %d, waiting for worker to finish...", pool_size)
                    serv_job.worker_complete.wait()
                    serv_job.worker_complete.clear()
                    # remove complete workers
                    LOG.debug("trimming worker pool")
                    # sometimes the thread that triggered the event doesn't quite cleanup in time
                    # so add a retry (10x with 0.5 second sleep on failure)
                    for _ in range(10):
                        worker_pool = list(w for w in worker_pool if not w.done)
                        pool_size = len(worker_pool)
                        if pool_size < max_workers:
                            break
                        time.sleep(0.5)  # pragma: no cover
                    else:  # pragma: no cover
                        # this should never happen
                        raise RuntimeError("Failed to trim worker pool!")
                    LOG.debug("trimmed worker pool (size: %d)", pool_size)
        except Exception:  # pylint: disable=broad-except
            if serv_job.exceptions.empty():
                serv_job.exceptions.put(sys.exc_info())
            serv_job.finish()
        finally:
            LOG.debug("listener cleaning up workers")
            deadline = time.time() + shutdown_delay
            while time.time() < deadline:
                worker_pool = list(w for w in worker_pool if not w.done)
                if not worker_pool:
                    break
                # avoid cutting off connections
                LOG.debug("waiting for %d worker(s)...", len(worker_pool))
                time.sleep(0.1)
            else:  # pragma: no cover
                LOG.debug("closing remaining workers")
                for worker in (w for w in worker_pool if not w.done):
                    worker.close()
